"""Storage paths snapshot."""

DATA_STORAGE_PATHS = {
    "count": 3,
    "next": None,
    "previous": None,
    "all": [1, 2, 3, 4, 5],
    "results": [
        {
            "id": 1,
            "slug": "work",
            "name": "Work Work Work",
            "path": "{owner_username}/work/{correspondent}_{created}_{document_type}_{title}",
            "match": "",
            "matching_algorithm": 6,
            "is_insensitive": True,
            "document_count": 384,
            "owner": 3,
            "user_can_change": True,
        },
        {
            "id": 2,
            "slug": "banking",
            "name": "Banking",
            "path": "{owner_username}/banking/{correspondent}_{created}_{document_type}_{title}",
            "match": "",
            "matching_algorithm": 6,
            "is_insensitive": True,
            "document_count": 303,
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 3,
            "slug": "another-test",
            "name": "Another Test",
            "path": "Test/Path/{doc_pk}",
            "match": "",
            "matching_algorithm": 0,
            "is_insensitive": True,
            "document_count": 0,
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 4,
            "slug": "another-test-2",
            "name": "Another Test 2",
            "path": "Test/Path/{doc_pk}",
            "match": "",
            "matching_algorithm": 0,
            "is_insensitive": True,
            "document_count": 0,
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 5,
            "slug": "another-test-3",
            "name": "Another Test 3",
            "path": "Test/Path/{doc_pk}",
            "match": "",
            "matching_algorithm": 0,
            "is_insensitive": True,
            "document_count": 0,
            "owner": None,
            "user_can_change": True,
        },
    ],
}



"""Custom fields snapshot."""

DATA_CUSTOM_FIELDS = {
    "count": 8,
    "next": None,
    "previous": None,
    "all": [8, 7, 6, 5, 4, 3, 2, 1],
    "results": [
        {"id": 8, "name": "Custom Link", "data_type": "documentlink"},
        {"id": 7, "name": "Custom URL", "data_type": "url"},
        {"id": 6, "name": "Custom Text -added-", "data_type": "string"},
        {"id": 5, "name": "Custom MONEYY $$$", "data_type": "monetary"},
        {"id": 4, "name": "Custom Floating", "data_type": "float"},
        {"id": 3, "name": "Custom Int", "data_type": "integer"},
        {"id": 2, "name": "Custom Date", "data_type": "date"},
        {"id": 1, "name": "Custom Bool", "data_type": "boolean"},
    ],
}



"""Document suggestions snapshot."""

DATA_DOCUMENT_SUGGESTIONS = {
    "correspondents": [26],
    "tags": [
        1,
        2,
        3,
    ],
    "document_types": [4],
    "storage_paths": [
        3,
        5,
    ],
    "dates": [
        "2022-01-07",
        "2023-01-07",
    ],
}



"""Tests for pypaperless."""

from dataclasses import dataclass
from typing import Any

from pypaperless import helpers, models
from pypaperless.const import PaperlessResource
from pypaperless.models import common

from .data import (
    DATA_CONFIG,
    DATA_CORRESPONDENTS,
    DATA_CUSTOM_FIELDS,
    DATA_DOCUMENT_TYPES,
    DATA_DOCUMENTS,
    DATA_GROUPS,
    DATA_MAIL_ACCOUNTS,
    DATA_MAIL_RULES,
    DATA_PROCESSED_MAIL,
    DATA_SAVED_VIEWS,
    DATA_SHARE_LINKS,
    DATA_STATUS,
    DATA_STORAGE_PATHS,
    DATA_TAGS,
    DATA_TASKS,
    DATA_USERS,
    DATA_WORKFLOWS,
)

# mypy: ignore-errors


@dataclass
class ResourceTestMapping:
    """Mapping for test cases."""

    resource: str
    data: dict[str, Any] | list[dict[str, Any]]
    helper_cls: type
    model_cls: type
    draft_cls: type | None = None
    draft_defaults: dict[str, Any] | None = None


CONFIG_MAP = ResourceTestMapping(
    PaperlessResource.CONFIG,
    DATA_CONFIG,
    helpers.ConfigHelper,
    models.Config,
)

CORRESPONDENT_MAP = ResourceTestMapping(
    PaperlessResource.CORRESPONDENTS,
    DATA_CORRESPONDENTS,
    helpers.CorrespondentHelper,
    models.Correspondent,
    models.CorrespondentDraft,
    {
        "name": "New Correspondent",
        "match": "",
        "matching_algorithm": common.MatchingAlgorithmType.ANY,
        "is_insensitive": True,
    },
)

CUSTOM_FIELD_MAP = ResourceTestMapping(
    PaperlessResource.CUSTOM_FIELDS,
    DATA_CUSTOM_FIELDS,
    helpers.CustomFieldHelper,
    models.CustomField,
    models.CustomFieldDraft,
    {
        "name": "New Custom Field",
        "data_type": common.CustomFieldType.BOOLEAN,
    },
)

DOCUMENT_MAP = ResourceTestMapping(
    PaperlessResource.DOCUMENTS,
    DATA_DOCUMENTS,
    helpers.DocumentHelper,
    models.Document,
    models.DocumentDraft,
    {
        "document": b"...example...content...",
        "tags": [1, 2, 3],
        "correspondent": 1,
        "document_type": 1,
        "storage_path": 1,
        "title": "New Document",
        "created": None,
        "archive_serial_number": 1,
    },
)

DOCUMENT_TYPE_MAP = ResourceTestMapping(
    PaperlessResource.DOCUMENT_TYPES,
    DATA_DOCUMENT_TYPES,
    helpers.DocumentTypeHelper,
    models.DocumentType,
    models.DocumentTypeDraft,
    {
        "name": "New Document Type",
        "match": "",
        "matching_algorithm": common.MatchingAlgorithmType.ANY,
        "is_insensitive": True,
    },
)

GROUP_MAP = ResourceTestMapping(
    PaperlessResource.GROUPS,
    DATA_GROUPS,
    helpers.GroupHelper,
    models.Group,
)

MAIL_ACCOUNT_MAP = ResourceTestMapping(
    PaperlessResource.MAIL_ACCOUNTS,
    DATA_MAIL_ACCOUNTS,
    helpers.MailAccountHelper,
    models.MailAccount,
)

MAIL_RULE_MAP = ResourceTestMapping(
    PaperlessResource.MAIL_RULES,
    DATA_MAIL_RULES,
    helpers.MailRuleHelper,
    models.MailRule,
)

PROCESSED_MAIL_MAP = ResourceTestMapping(
    PaperlessResource.PROCESSED_MAIL,
    DATA_PROCESSED_MAIL,
    helpers.ProcessedMailHelper,
    models.ProcessedMail,
)

SAVED_VIEW_MAP = ResourceTestMapping(
    PaperlessResource.SAVED_VIEWS,
    DATA_SAVED_VIEWS,
    helpers.SavedViewHelper,
    models.SavedView,
)

SHARE_LINK_MAP = ResourceTestMapping(
    PaperlessResource.SHARE_LINKS,
    DATA_SHARE_LINKS,
    helpers.ShareLinkHelper,
    models.ShareLink,
    models.ShareLinkDraft,
    {
        "expiration": None,
        "document": 1,
        "file_version": common.ShareLinkFileVersionType.ORIGINAL,
    },
)

STATUS_MAP = ResourceTestMapping(
    PaperlessResource.STATUS,
    DATA_STATUS,
    helpers.StatusHelper,
    models.Status,
)

STORAGE_PATH_MAP = ResourceTestMapping(
    PaperlessResource.STORAGE_PATHS,
    DATA_STORAGE_PATHS,
    helpers.StoragePathHelper,
    models.StoragePath,
    models.StoragePathDraft,
    {
        "name": "New Storage Path",
        "path": "path/to/test",
        "match": "",
        "matching_algorithm": common.MatchingAlgorithmType.ANY,
        "is_insensitive": True,
    },
)

TAG_MAP = ResourceTestMapping(
    PaperlessResource.TAGS,
    DATA_TAGS,
    helpers.TagHelper,
    models.Tag,
    models.TagDraft,
    {
        "name": "New Tag",
        "color": "#012345",
        "text_color": "#987654",
        "is_inbox_tag": False,
        "match": "",
        "matching_algorithm": common.MatchingAlgorithmType.ANY,
        "is_insensitive": True,
    },
)

TASK_MAP = ResourceTestMapping(
    PaperlessResource.TASKS,
    DATA_TASKS,
    helpers.TaskHelper,
    models.Task,
)

USER_MAP = ResourceTestMapping(
    PaperlessResource.USERS,
    DATA_USERS,
    helpers.UserHelper,
    models.User,
)

WORKFLOW_MAP = ResourceTestMapping(
    PaperlessResource.WORKFLOWS,
    DATA_WORKFLOWS,
    helpers.WorkflowHelper,
    models.Workflow,
)



"""Object permissions snapshot."""

DATA_OBJECT_PERMISSIONS = {
    "view": {
        "users": [1, 2],
        "groups": [],
    },
    "change": {
        "users": [],
        "groups": [1],
    },
}



"""Workflow snapshot."""

# mypy: ignore-errors

DATA_WORKFLOWS = {
    "count": 3,
    "next": None,
    "previous": None,
    "all": [1, 2, 3],
    "results": [
        {
            "id": 1,
            "name": "Importordner Template",
            "order": 1,
            "enabled": True,
            "triggers": [
                {
                    "id": 1,
                    "sources": [1],
                    "type": 1,
                    "filter_path": None,
                    "filter_filename": "*.pdf",
                    "filter_mailrule": None,
                    "matching_algorithm": 0,
                    "match": "",
                    "is_insensitive": None,
                    "filter_has_tags": [],
                    "filter_has_correspondent": None,
                    "filter_has_document_type": None,
                }
            ],
            "actions": [
                {
                    "id": 1,
                    "type": 1,
                    "assign_title": "Some workflow title",
                    "assign_tags": [4],
                    "assign_correspondent": 9,
                    "assign_document_type": 8,
                    "assign_storage_path": 2,
                    "assign_owner": 3,
                    "assign_view_users": [],
                    "assign_view_groups": [],
                    "assign_change_users": [],
                    "assign_change_groups": [],
                    "assign_custom_fields": [2],
                }
            ],
        },
        {
            "id": 2,
            "name": "API Upload Template",
            "order": 2,
            "enabled": True,
            "triggers": [
                {
                    "id": 2,
                    "sources": [1, 2],
                    "type": 1,
                    "filter_path": "/api/*",
                    "filter_filename": "*.pdf",
                    "filter_mailrule": None,
                    "matching_algorithm": 0,
                    "match": "",
                    "is_insensitive": None,
                    "filter_has_tags": [],
                    "filter_has_correspondent": None,
                    "filter_has_document_type": None,
                }
            ],
            "actions": [
                {
                    "id": 2,
                    "type": 1,
                    "assign_title": "API",
                    "assign_tags": [4],
                    "assign_correspondent": 9,
                    "assign_document_type": 12,
                    "assign_storage_path": 2,
                    "assign_owner": 3,
                    "assign_view_users": [],
                    "assign_view_groups": [],
                    "assign_change_users": [],
                    "assign_change_groups": [],
                    "assign_custom_fields": [5],
                }
            ],
        },
        {
            "id": 3,
            "name": "Email Template",
            "order": 3,
            "enabled": False,
            "triggers": [
                {
                    "id": 3,
                    "sources": [3],
                    "type": 1,
                    "filter_path": "/mail/*",
                    "filter_filename": "*.eml",
                    "filter_mailrule": 1,
                    "matching_algorithm": 0,
                    "match": "",
                    "is_insensitive": True,
                    "filter_has_tags": [],
                    "filter_has_correspondent": None,
                    "filter_has_document_type": None,
                }
            ],
            "actions": [
                {
                    "id": 3,
                    "type": 1,
                    "assign_title": None,
                    "assign_tags": [],
                    "assign_correspondent": None,
                    "assign_document_type": None,
                    "assign_storage_path": None,
                    "assign_owner": 2,
                    "assign_view_users": [3, 7],
                    "assign_view_groups": [1],
                    "assign_change_users": [6],
                    "assign_change_groups": [],
                    "assign_custom_fields": [],
                }
            ],
        },
    ],
}

DATA_WORKFLOW_ACTIONS = {
    "count": 0,
    "next": None,
    "previous": None,
    "all": [],
    "results": [],
}
for wf in DATA_WORKFLOWS["results"]:
    DATA_WORKFLOW_ACTIONS["count"] += 1
    for act in wf["actions"]:
        DATA_WORKFLOW_ACTIONS["all"].append(act["id"])
        DATA_WORKFLOW_ACTIONS["results"].append(act)

DATA_WORKFLOW_TRIGGERS = {
    "count": 0,
    "next": None,
    "previous": None,
    "all": [],
    "results": [],
}
for wf in DATA_WORKFLOWS["results"]:
    DATA_WORKFLOW_TRIGGERS["count"] += 1
    for act in wf["triggers"]:
        DATA_WORKFLOW_TRIGGERS["all"].append(act["id"])
        DATA_WORKFLOW_TRIGGERS["results"].append(act)



"""Setup pytest."""

from collections.abc import AsyncGenerator, Generator
from typing import Any

import pytest
from aioresponses import aioresponses

from pypaperless import Paperless
from pypaperless.const import API_PATH

from .const import PAPERLESS_TEST_REQ_ARGS, PAPERLESS_TEST_TOKEN, PAPERLESS_TEST_URL
from .data import DATA_SCHEMA

# mypy: ignore-errors


@pytest.fixture(name="resp")
def aioresponses_fixture() -> Generator[aioresponses]:
    """Return aioresponses fixture."""
    with aioresponses() as m:
        yield m


@pytest.fixture(name="api")
def api_obj_fixture() -> Paperless:
    """Return Paperless."""
    return Paperless(
        PAPERLESS_TEST_URL,
        PAPERLESS_TEST_TOKEN,
        request_args=PAPERLESS_TEST_REQ_ARGS,
    )


@pytest.fixture(name="paperless")
async def paperless_fixture(
    resp: aioresponses,
    api: Paperless,
) -> AsyncGenerator[Paperless, Any]:
    """Return a Paperless object with given version."""
    resp.get(
        f"{PAPERLESS_TEST_URL}{API_PATH['index']}",
        status=200,
        payload=DATA_SCHEMA,
    )
    async with api:
        yield api



"""Paperless basic tests."""

import re
from typing import Any

import aiohttp
import pytest
from aioresponses import CallbackResult, aioresponses

from pypaperless import Paperless
from pypaperless.const import API_PATH
from pypaperless.exceptions import DraftFieldRequiredError
from pypaperless.models import Page
from pypaperless.models.common import PermissionTableType

from . import (
    CORRESPONDENT_MAP,
    CUSTOM_FIELD_MAP,
    DOCUMENT_MAP,
    DOCUMENT_TYPE_MAP,
    GROUP_MAP,
    MAIL_ACCOUNT_MAP,
    MAIL_RULE_MAP,
    PROCESSED_MAIL_MAP,
    SAVED_VIEW_MAP,
    SHARE_LINK_MAP,
    STORAGE_PATH_MAP,
    TAG_MAP,
    USER_MAP,
    WORKFLOW_MAP,
    ResourceTestMapping,
)
from .const import PAPERLESS_TEST_URL
from .data import DATA_OBJECT_PERMISSIONS

# mypy: ignore-errors


@pytest.mark.parametrize(
    "mapping",
    [
        DOCUMENT_MAP,
        DOCUMENT_TYPE_MAP,
        CORRESPONDENT_MAP,
        CUSTOM_FIELD_MAP,
        GROUP_MAP,
        MAIL_ACCOUNT_MAP,
        MAIL_RULE_MAP,
        PROCESSED_MAIL_MAP,
        SAVED_VIEW_MAP,
        SHARE_LINK_MAP,
        STORAGE_PATH_MAP,
        TAG_MAP,
        USER_MAP,
        WORKFLOW_MAP,
    ],
    scope="class",
)
# test models/classifiers.py
# test models/custom_fields.py
# test models/mails.py
# test models/permissions.py
# test models/saved_views.py
# test models/share_links.py
class TestReadOnly:
    """Read only resources test cases."""

    async def test_pages(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test pages."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload=mapping.data,
        )
        page = await anext(aiter(getattr(paperless, mapping.resource).pages(1)))
        assert isinstance(page, Page)
        assert isinstance(page.items, list)
        for item in page.items:
            assert isinstance(item, mapping.model_cls)

    async def test_as_dict(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test as_dict."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload=mapping.data,
        )
        items = await getattr(paperless, mapping.resource).as_dict()
        for pk, obj in items.items():
            assert isinstance(pk, int)
            assert isinstance(obj, mapping.model_cls)

    async def test_as_list(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test as_dict."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload=mapping.data,
        )
        items = await getattr(paperless, mapping.resource).as_list()
        for obj in items:
            assert isinstance(obj, mapping.model_cls)

    async def test_iter(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test iter."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload=mapping.data,
        )
        async for item in getattr(paperless, mapping.resource):
            assert isinstance(item, mapping.model_cls)

    async def test_all(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test all."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload=mapping.data,
        )
        items = await getattr(paperless, mapping.resource).all()
        assert isinstance(items, list)
        for item in items:
            assert isinstance(item, int)

    async def test_call(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test call."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1),
            status=200,
            payload=mapping.data["results"][0],
        )
        item = await getattr(paperless, mapping.resource)(1)
        assert item
        assert isinstance(item, mapping.model_cls)
        # must raise as 1337 doesn't exist
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1337),
            status=404,
        )
        with pytest.raises(aiohttp.ClientResponseError):
            await getattr(paperless, mapping.resource)(1337)


@pytest.mark.parametrize(
    "mapping",
    [
        CORRESPONDENT_MAP,
        CUSTOM_FIELD_MAP,
        DOCUMENT_TYPE_MAP,
        SHARE_LINK_MAP,
        STORAGE_PATH_MAP,
        TAG_MAP,
    ],
    scope="class",
)
# test models/classifiers.py
# test models/custom_fields.py
# test models/share_links.py
class TestReadWrite:
    """R/W models test cases."""

    async def test_pages(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test pages."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload=mapping.data,
        )
        page = await anext(aiter(getattr(paperless, mapping.resource).pages(1)))
        assert isinstance(page, Page)
        assert isinstance(page.items, list)
        for item in page.items:
            assert isinstance(item, mapping.model_cls)

    async def test_iter(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test iter."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload=mapping.data,
        )
        async for item in getattr(paperless, mapping.resource):
            assert isinstance(item, mapping.model_cls)

    async def test_all(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test all."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload=mapping.data,
        )
        items = await getattr(paperless, mapping.resource).all()
        assert isinstance(items, list)
        for item in items:
            assert isinstance(item, int)

    async def test_reduce(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test iter with reduce."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload=mapping.data,
        )
        async with getattr(paperless, mapping.resource).reduce(
            any_filter_param="1",
            any_filter_list__in=["1", "2"],
            any_filter_no_list__in="1",
        ) as q:
            async for item in q:
                assert isinstance(item, mapping.model_cls)

    async def test_call(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test call."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1),
            status=200,
            payload=mapping.data["results"][0],
        )
        item = await getattr(paperless, mapping.resource)(1)
        assert item
        assert isinstance(item, mapping.model_cls)
        # must raise as 1337 doesn't exist
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1337),
            status=404,
        )
        with pytest.raises(aiohttp.ClientResponseError):
            await getattr(paperless, mapping.resource)(1337)

    async def test_create(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test create."""
        draft = getattr(paperless, mapping.resource).draft(**mapping.draft_defaults)
        assert isinstance(draft, mapping.draft_cls)
        # test empty draft fields
        if mapping.model_cls not in (
            SHARE_LINK_MAP.model_cls,
            CUSTOM_FIELD_MAP.model_cls,
        ):
            backup = draft.name
            draft.name = None
            with pytest.raises(DraftFieldRequiredError):
                await draft.save()
            draft.name = backup
        # actually call the create endpoint
        resp.post(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}",
            status=200,
            payload={
                "id": len(mapping.data["results"]),
                **draft._serialize(),  # pylint: disable=protected-access
            },
        )
        new_pk = await draft.save()
        assert new_pk >= 1

    async def test_udpate(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test update."""
        update_field = "name"
        update_value = "Name Updated"
        if mapping.model_cls is SHARE_LINK_MAP.model_cls:
            update_field = "document"
            update_value = 2
        # go on
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1),
            status=200,
            payload=mapping.data["results"][0],
        )
        to_update = await getattr(paperless, mapping.resource)(1)
        setattr(to_update, update_field, update_value)
        # actually call the update endpoint
        resp.patch(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1),
            status=200,
            payload={
                **to_update._data,  # pylint: disable=protected-access
                update_field: update_value,
            },
        )
        await to_update.update()
        assert getattr(to_update, update_field) == update_value
        # no updates
        assert not await to_update.update()
        # force update
        setattr(to_update, update_field, update_value)
        resp.put(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1),
            status=200,
            payload={
                **to_update._data,  # pylint: disable=protected-access
                update_field: update_value,
            },
        )
        await to_update.update(only_changed=False)
        assert getattr(to_update, update_field) == update_value

    async def test_delete(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test delete."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1),
            status=200,
            payload=mapping.data["results"][0],
        )
        to_delete = await getattr(paperless, mapping.resource)(1)
        resp.delete(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1),
            status=204,  # Paperless-ngx responds with 204 on deletion
        )
        assert await to_delete.delete()
        # test deletion failed
        resp.delete(
            f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1),
            status=404,  # we send another status code
        )
        assert not await to_delete.delete()


@pytest.mark.parametrize(
    "mapping",
    [
        CORRESPONDENT_MAP,
        DOCUMENT_MAP,
        DOCUMENT_TYPE_MAP,
        STORAGE_PATH_MAP,
        TAG_MAP,
    ],
    scope="class",
)
# test models/classifiers.py
class TestSecurableMixin:
    """SecurableMixin test cases."""

    async def test_permissions(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test permissions."""
        getattr(paperless, mapping.resource).request_permissions = True
        assert getattr(paperless, mapping.resource).request_permissions
        # request single object
        resp.get(
            re.compile(
                r"^"
                + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1)
                + r"\?.*$"
            ),
            status=200,
            payload={
                **mapping.data["results"][0],
                "permissions": DATA_OBJECT_PERMISSIONS,
            },
        )
        item = await getattr(paperless, mapping.resource)(1)
        assert item.has_permissions
        assert isinstance(item.permissions, PermissionTableType)
        # request by iterator
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource]}" + r"\?.*$"),
            status=200,
            payload={
                **mapping.data,
                "results": [
                    {**item, "permissions": DATA_OBJECT_PERMISSIONS}
                    for item in mapping.data["results"]
                ],
            },
        )
        async for item in getattr(paperless, mapping.resource):
            assert isinstance(item, mapping.model_cls)
            assert item.has_permissions
            assert isinstance(item.permissions, PermissionTableType)

    async def test_permission_change(
        self, resp: aioresponses, paperless: Paperless, mapping: ResourceTestMapping
    ) -> None:
        """Test permission changes."""
        getattr(paperless, mapping.resource).request_permissions = True
        assert getattr(paperless, mapping.resource).request_permissions
        resp.get(
            re.compile(
                r"^"
                + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1)
                + r"\?.*$"
            ),
            status=200,
            payload={
                **mapping.data["results"][0],
                "permissions": DATA_OBJECT_PERMISSIONS,
            },
        )
        item = await getattr(paperless, mapping.resource)(1)
        item.permissions.view.users.append(23)

        def _lookup_set_permissions(  # pylint: disable=unused-argument
            url: str,
            json: dict[str, Any],
            **kwargs: Any,  # noqa: ARG001
        ) -> CallbackResult:
            assert url
            assert "set_permissions" in json
            return CallbackResult(
                status=200,
                payload=item._data,  # pylint: disable=protected-access
            )

        resp.patch(
            re.compile(
                r"^"
                + f"{PAPERLESS_TEST_URL}{API_PATH[mapping.resource + '_single']}".format(pk=1)
                + r"\?.*$"
            ),
            callback=_lookup_set_permissions,
        )
        await item.update()



"""PyPaperless common types."""

import contextlib
import datetime
import re
from dataclasses import dataclass, field
from enum import Enum, StrEnum
from typing import TYPE_CHECKING, Any, TypedDict, TypeVar

if TYPE_CHECKING:
    from pypaperless import Paperless

    from .classifiers import Correspondent, DocumentType, StoragePath, Tag
    from .custom_fields import CustomField


# custom_fields
class CustomFieldExtraDataSelectOptions(TypedDict):
    """Represent the `extra_data.select_options` field of a `CustomField`."""

    id: str | None
    label: str | None


class CustomFieldExtraData(TypedDict):
    """Represent the `extra_data` field of a `CustomField`."""

    default_currency: str | None
    select_options: list[CustomFieldExtraDataSelectOptions | None]


class CustomFieldType(Enum):
    """Represent a subtype of `CustomField`."""

    STRING = "string"
    URL = "url"
    DATE = "date"
    BOOLEAN = "boolean"
    INTEGER = "integer"
    FLOAT = "float"
    MONETARY = "monetary"
    DOCUMENT_LINK = "documentlink"
    SELECT = "select"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: type, *_: object) -> "CustomFieldType":
        """Set default member on unknown value."""
        return CustomFieldType.UNKNOWN


# documents
@dataclass(kw_only=True)
class CustomFieldValue:
    """Represent a subtype of `CustomField`."""

    field: int | None = None
    value: Any | None = None
    name: str | None = None
    data_type: CustomFieldType | None = None
    extra_data: CustomFieldExtraData | None = None


CustomFieldValueT = TypeVar("CustomFieldValueT", bound=CustomFieldValue)


@dataclass(kw_only=True)
class CustomFieldBooleanValue(CustomFieldValue):
    """Represent a boolean `CustomFieldValue`."""

    value: bool | None = None


@dataclass(kw_only=True)
class CustomFieldDateValue(CustomFieldValue):
    """Represent a date `CustomFieldValue`."""

    value: datetime.date | str | None = None

    def __post_init__(self) -> None:
        """Convert the value to a datetime."""
        if isinstance(self.value, str):
            with contextlib.suppress(ValueError):
                dt = datetime.datetime.fromisoformat(self.value)
                self.value = dt.date()


@dataclass(kw_only=True)
class CustomFieldDocumentLinkValue(CustomFieldValue):
    """Represent a document link `CustomFieldValue`."""

    value: list[int] | None = None


@dataclass(kw_only=True)
class CustomFieldFloatValue(CustomFieldValue):
    """Represent a float `CustomFieldValue`."""

    value: float | None = None


@dataclass(kw_only=True)
class CustomFieldIntegerValue(CustomFieldValue):
    """Represent an integer `CustomFieldValue`."""

    value: int | None = None


@dataclass(kw_only=True)
class CustomFieldMonetaryValue(CustomFieldValue):
    """Represent a monetary `CustomFieldValue`."""

    value: str | None = None

    @property
    def currency(self) -> str | None:
        """Return the currency of the `value` field."""
        if self.value and (match := re.match(r"^([a-zA-Z]{3})", self.value)):
            return match.group(1) if match else self.value
        if self.extra_data and (default_currency := self.extra_data.get("default_currency", None)):
            return default_currency
        return ""

    @currency.setter
    def currency(self, new_currency: str) -> None:
        """Override the currency of the field."""
        value = self.value or ""
        value = re.sub(r"^[a-zA-Z]{3}", "", value)

        if new_currency and re.match(r"^[A-Z]{3}$", new_currency):
            self.value = f"{new_currency}{value}"
        else:
            self.value = value

    @property
    def amount(self) -> float | None:
        """Return the amount without currentcy of the `value` field."""
        if self.value:
            numeric = re.sub(r"[^\d.]", "", self.value)
            return float(numeric)
        return None

    @amount.setter
    def amount(self, new_amount: float) -> None:
        """Set a new amount and construct the internal needed value."""
        self.value = f"{self.currency}{new_amount:.2f}"


@dataclass(kw_only=True)
class CustomFieldSelectValue(CustomFieldValue):
    """Represent a select `CustomFieldValue`."""

    value: int | str | None = None

    @property
    def labels(self) -> list[CustomFieldExtraDataSelectOptions | None]:
        """Return the list of labels of the `CustomField`."""
        if not self.extra_data:
            return []
        return self.extra_data["select_options"]

    @property
    def label(self) -> str | None:
        """Return the label for `value` or fall back to `None`."""
        for opt in self.labels:
            if opt and opt["id"] == self.value:
                return opt["label"]
        return None


@dataclass(kw_only=True)
class CustomFieldStringValue(CustomFieldValue):
    """Represent a string `CustomFieldValue`."""

    value: str | None = None


@dataclass(kw_only=True)
class CustomFieldURLValue(CustomFieldValue):
    """Represent an url `CustomFieldValue`."""

    value: str | None = None


CUSTOM_FIELD_TYPE_VALUE_MAP: dict[CustomFieldType, type[CustomFieldValue]] = {
    CustomFieldType.BOOLEAN: CustomFieldBooleanValue,
    CustomFieldType.DATE: CustomFieldDateValue,
    CustomFieldType.DOCUMENT_LINK: CustomFieldDocumentLinkValue,
    CustomFieldType.FLOAT: CustomFieldFloatValue,
    CustomFieldType.INTEGER: CustomFieldIntegerValue,
    CustomFieldType.MONETARY: CustomFieldMonetaryValue,
    CustomFieldType.SELECT: CustomFieldSelectValue,
    CustomFieldType.STRING: CustomFieldStringValue,
    CustomFieldType.URL: CustomFieldURLValue,
}


# documents
@dataclass(kw_only=True)
class DocumentMetadataType:
    """Represent a subtype of `DocumentMeta`."""

    namespace: str | None = None
    prefix: str | None = None
    key: str | None = None
    value: str | None = None


# documents
@dataclass(kw_only=True)
class DocumentSearchHitType:
    """Represent a subtype of `Document`."""

    score: float | None = None
    highlights: str | None = None
    note_highlights: str | None = None
    rank: int | None = None


# api
@dataclass(kw_only=True)
class MasterDataInstance:
    """Represent a `MasterDataInstance`."""

    api: "Paperless"
    is_initialized: bool = False

    correspondents: list["Correspondent"] = field(default_factory=list)
    custom_fields: list["CustomField"] = field(default_factory=list)
    document_types: list["DocumentType"] = field(default_factory=list)
    storage_paths: list["StoragePath"] = field(default_factory=list)
    tags: list["Tag"] = field(default_factory=list)


# mixins/models/data_fields, used for classifiers
class MatchingAlgorithmType(Enum):
    """Represent a subtype of `Correspondent`, `DocumentType`, `StoragePath` and `Tag`."""

    NONE = 0
    ANY = 1
    ALL = 2
    LITERAL = 3
    REGEX = 4
    FUZZY = 5
    AUTO = 6
    UNKNOWN = -1

    @classmethod
    def _missing_(cls: type, *_: object) -> "MatchingAlgorithmType":
        """Set default member on unknown value."""
        return MatchingAlgorithmType.UNKNOWN


# api
@dataclass(kw_only=True)
class PaperlessCache:
    """Represent a Paperless cache object."""

    custom_fields: dict[int, "CustomField"] | None = None


# mixins/models/securable
@dataclass(kw_only=True)
class PermissionSetType:
    """Represent a Paperless permission set."""

    users: list[int] = field(default_factory=list)
    groups: list[int] = field(default_factory=list)


# mixins/models/securable
@dataclass(kw_only=True)
class PermissionTableType:
    """Represent a Paperless permissions type."""

    view: PermissionSetType = field(default_factory=PermissionSetType)
    change: PermissionSetType = field(default_factory=PermissionSetType)


# documents
class RetrieveFileMode(StrEnum):
    """Represent a subtype of `DownloadedDocument`."""

    DOWNLOAD = "download"
    PREVIEW = "preview"
    THUMBNAIL = "thumb"


# saved_views
@dataclass(kw_only=True)
class SavedViewFilterRuleType:
    """Represent a subtype of `SavedView`."""

    rule_type: int | None = None
    value: str | None = None


# share_links
class ShareLinkFileVersionType(Enum):
    """Represent a subtype of `ShareLink`."""

    ARCHIVE = "archive"
    ORIGINAL = "original"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: type, *_: object) -> "ShareLinkFileVersionType":
        """Set default member on unknown value."""
        return ShareLinkFileVersionType.UNKNOWN


# statistics
@dataclass(kw_only=True)
class StatisticDocumentFileTypeCount:
    """Represent a Paperless statistics file type count."""

    mime_type: str | None = None
    mime_type_count: int | None = None


# status
class StatusType(Enum):
    """Represent a subtype of `Status`."""

    OK = "OK"
    ERROR = "ERROR"
    WARNING = "WARNING"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def _missing_(cls: type, *_: object) -> "StatusType":
        """Set default member on unknown value."""
        return StatusType.UNKNOWN


# status
@dataclass(kw_only=True)
class StatusDatabaseMigrationStatusType:
    """Represent a subtype of `StatusDatabaseType`."""

    latest_migration: str | None = None
    unapplied_migrations: list[str] = field(default_factory=list)


# status
@dataclass(kw_only=True)
class StatusDatabaseType:
    """Represent a subtype of `Status`."""

    type: str | None = None
    url: str | None = None
    status: StatusType | None = None
    error: str | None = None
    migration_status: StatusDatabaseMigrationStatusType | None = None


# status
@dataclass(kw_only=True)
class StatusStorageType:
    """Represent a subtype of `Status`."""

    total: int | None = None
    available: int | None = None


# status
@dataclass(kw_only=True)
class StatusTasksType:
    """Represent a subtype of `Status`."""

    redis_url: str | None = None
    redis_status: StatusType | None = None
    redis_error: str | None = None
    celery_status: StatusType | None = None
    celery_url: str | None = None
    celery_error: str | None = None
    index_status: StatusType | None = None
    index_last_modified: datetime.datetime | None = None
    index_error: str | None = None
    classifier_status: StatusType | None = None
    classifier_last_trained: datetime.datetime | None = None
    classifier_error: str | None = None
    sanity_check_status: StatusType | None = None
    sanity_check_last_run: datetime.datetime | None = None
    sanity_check_error: str | None = None


# tasks
class TaskType(Enum):
    """Represent a subtype of `Task`."""

    AUTO = "auto_task"
    SCHEDULED = "scheduled_task"
    MANUAL = "manual_task"
    UNKNOWN = "unknown"

    @classmethod
    def _missing_(cls: type, *_: object) -> "TaskType":
        """Set default member on unknown value."""
        return TaskType.UNKNOWN


# tasks
class TaskStatusType(Enum):
    """Represent a subtype of `Task`."""

    FAILURE = "FAILURE"
    PENDING = "PENDING"
    RECEIVED = "RECEIVED"
    RETRY = "RETRY"
    REVOKED = "REVOKED"
    STARTED = "STARTED"
    SUCCESS = "SUCCESS"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def _missing_(cls: type, *_: object) -> "TaskStatusType":
        """Set default member on unknown value."""
        return TaskStatusType.UNKNOWN


# workflows
class WorkflowActionType(Enum):
    """Represent a subtype of `Workflow`."""

    ASSIGNMENT = 1
    REMOVAL = 2
    EMAIL = 3
    WEBHOOK = 4
    UNKNOWN = -1

    @classmethod
    def _missing_(cls: type, *_: object) -> "WorkflowActionType":
        """Set default member on unknown value."""
        return WorkflowActionType.UNKNOWN


# workflows
@dataclass(kw_only=True)
class WorkflowActionEmailType:
    """Represent a subtype of `WorkflowAction`."""

    id: int | None = None
    to: str | None = None
    subject: str | None = None
    body: str | None = None
    include_document: bool | None = None


@dataclass(kw_only=True)
class WorkflowActionWebhookType:
    """Represent a subtype of `WorkflowAction`."""

    id: int | None = None
    url: str | None = None
    use_params: bool | None = None
    as_json: bool | None = None
    params: dict[str, Any] | None = None
    body: str | None = None
    headers: dict[str, str] | None = None
    include_document: bool | None = None


# workflows
class WorkflowTriggerType(Enum):
    """Represent a subtype of `Workflow`."""

    CONSUMPTION = 1
    DOCUMENT_ADDED = 2
    DOCUMENT_UPDATED = 3
    SCHEDULED = 4
    UNKNOWN = -1

    @classmethod
    def _missing_(cls: type, *_: object) -> "WorkflowTriggerType":
        """Set default member on unknown value."""
        return WorkflowTriggerType.UNKNOWN


# workflows
class WorkflowTriggerScheduleDateFieldType(Enum):
    """Represent a subtype of `WorkflowTrigger`."""

    ADDED = "added"
    CREATED = "created"
    MODIFIED = "modified"
    CUSTOM_FIELD = "custom_field"
    UNKNOWN = -1

    @classmethod
    def _missing_(cls: type, *_: object) -> "WorkflowTriggerScheduleDateFieldType":
        """Set default member on unknown value."""
        return WorkflowTriggerScheduleDateFieldType.UNKNOWN


# workflows
class WorkflowTriggerSourceType(Enum):
    """Represent a subtype of `Workflow`."""

    CONSUME_FOLDER = 1
    API_UPLOAD = 2
    MAIL_FETCH = 3
    UNKNOWN = -1

    @classmethod
    def _missing_(cls: type, *_: object) -> "WorkflowTriggerSourceType":
        """Set default member on unknown value."""
        return WorkflowTriggerSourceType.UNKNOWN



"""Test constants."""

PAPERLESS_TEST_URL = "https://paperless.not-existing.internal"
PAPERLESS_TEST_TOKEN = "abcdef0123456789"
PAPERLESS_TEST_USER = "test-user"
PAPERLESS_TEST_PASSWORD = "not-so-secret-password"
PAPERLESS_TEST_REQ_ARGS = {"ssl": False}



"""Provide `CustomField` related models and helpers."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, overload

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .common import (
    CUSTOM_FIELD_TYPE_VALUE_MAP,
    CustomFieldExtraData,
    CustomFieldType,
    CustomFieldValue,
    CustomFieldValueT,
)
from .mixins import helpers, models

if TYPE_CHECKING:
    from pypaperless import Paperless


@dataclass(init=False)
class CustomField(
    PaperlessModel,
    models.UpdatableMixin,
    models.DeletableMixin,
):
    """Represent a Paperless `CustomField`."""

    _api_path = API_PATH["custom_fields_single"]

    id: int
    name: str | None = None
    data_type: CustomFieldType | None = None
    extra_data: CustomFieldExtraData | None = None
    document_count: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Document` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))

    @overload
    def draft_value(self, value: Any) -> CustomFieldValue: ...

    @overload
    def draft_value(
        self, value: Any, expected_type: type[CustomFieldValueT]
    ) -> CustomFieldValueT: ...

    def draft_value(
        self,
        value: Any,
        expected_type: type[CustomFieldValueT] | None = None,  # noqa: ARG002 # pylint: disable=unused-argument
    ) -> CustomFieldValue | CustomFieldValueT:
        """Draft a new `CustomFieldValue` instance."""
        cache = self._api.cache.custom_fields

        if cache and self.id in cache:
            klass = CUSTOM_FIELD_TYPE_VALUE_MAP.get(
                self.data_type or CustomFieldType.UNKNOWN, CustomFieldValue
            )
            klass_data = {
                "field": self.id,
                "value": value,
                "name": self.name,
                "data_type": self.data_type,
                "extra_data": self.extra_data,
            }
            return klass(**klass_data)

        return CustomFieldValue(field=self.id, value=value)


@dataclass(init=False)
class CustomFieldDraft(PaperlessModel, models.CreatableMixin):
    """Represent a new Paperless `CustomField`, which is not stored in Paperless."""

    _api_path = API_PATH["custom_fields"]

    _create_required_fields = {"name", "data_type"}

    name: str | None = None
    data_type: CustomFieldType | None = None
    extra_data: CustomFieldExtraData | None = None


class CustomFieldHelper(
    HelperBase,
    helpers.CallableMixin[CustomField],
    helpers.DraftableMixin[CustomFieldDraft],
    helpers.IterableMixin[CustomField],
):
    """Represent a factory for Paperless `CustomField` models."""

    _api_path = API_PATH["custom_fields"]
    _resource = PaperlessResource.CUSTOM_FIELDS

    _draft_cls = CustomFieldDraft
    _resource_cls = CustomField



"""Paperless common tests."""

import re
from dataclasses import dataclass, fields
from datetime import date, datetime
from enum import Enum
from typing import TypedDict

import aiohttp
import pytest
from aioresponses import aioresponses

from pypaperless import Paperless
from pypaperless.const import API_PATH, PaperlessResource
from pypaperless.exceptions import (
    BadJsonResponseError,
    DraftNotSupportedError,
    InitializationError,
    JsonResponseWithError,
    PaperlessConnectionError,
    PaperlessForbiddenError,
    PaperlessInactiveOrDeletedError,
    PaperlessInvalidTokenError,
)
from pypaperless.models import CustomField, Page
from pypaperless.models.base import HelperBase, PaperlessModel
from pypaperless.models.common import (
    CUSTOM_FIELD_TYPE_VALUE_MAP,
    CustomFieldDateValue,
    CustomFieldIntegerValue,
    CustomFieldMonetaryValue,
    CustomFieldSelectValue,
    CustomFieldType,
    MatchingAlgorithmType,
    ShareLinkFileVersionType,
    StatusType,
    TaskStatusType,
    TaskType,
    WorkflowActionType,
    WorkflowTriggerScheduleDateFieldType,
    WorkflowTriggerSourceType,
    WorkflowTriggerType,
)
from pypaperless.models.mixins import helpers
from pypaperless.models.utils import dict_value_to_object, object_to_dict_value
from tests.const import (
    PAPERLESS_TEST_PASSWORD,
    PAPERLESS_TEST_TOKEN,
    PAPERLESS_TEST_URL,
    PAPERLESS_TEST_USER,
)

from .data import DATA_CUSTOM_FIELDS, DATA_PATHS, DATA_TOKEN

# mypy: ignore-errors


class TestPaperless:
    """Paperless common test cases."""

    async def test_init(self, resp: aioresponses, api: Paperless) -> None:
        """Test init."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['index']}",
            status=200,
            payload=DATA_PATHS,
        )
        await api.initialize()
        assert api.is_initialized
        await api.close()

    async def test_context(self, resp: aioresponses, api: Paperless) -> None:
        """Test context."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['index']}",
            status=200,
            payload=DATA_PATHS,
        )
        async with api:
            assert api.is_initialized

    async def test_properties(self, api: Paperless) -> None:
        """Test properties."""
        # version must be None in this case, as we test against
        # an uninitialized Paperless object
        assert api.host_version is None
        assert api.base_url == PAPERLESS_TEST_URL

    async def test_init_error(self, resp: aioresponses, api: Paperless) -> None:
        """Test initialization error."""
        # simulate connection due no configuration error
        with pytest.raises(PaperlessConnectionError):
            await api.initialize()

        # http status error
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['index']}",
            status=401,
            body="any html",
        )
        with pytest.raises(PaperlessInvalidTokenError):
            await api.initialize()

        # http 401 - inactive or deleted user
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['index']}",
            status=401,
            payload={"detail": "User is inactive"},
        )
        with pytest.raises(PaperlessInactiveOrDeletedError):
            await api.initialize()

        # http status forbidden
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['index']}",
            status=403,
            body="any html",
        )
        with pytest.raises(PaperlessForbiddenError):
            await api.initialize()

        # http ok, wrong payload
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['index']}",
            status=200,
            body="any html",
        )
        with pytest.raises(InitializationError):
            await api.initialize()

    @pytest.mark.parametrize(
        "exception_cls",
        [
            PaperlessConnectionError,
            PaperlessInvalidTokenError,
            PaperlessInactiveOrDeletedError,
            PaperlessForbiddenError,
        ],
    )
    async def test_errors_are_backwards_compatible(self, exception_cls: type) -> None:
        """Test, if new errors are backwards compatible."""
        assert issubclass(exception_cls, InitializationError)

    async def test_jsonresponsewitherror(self) -> None:
        """Test JsonResponseWithError."""
        try:
            payload = "sample string"
            raise JsonResponseWithError(payload)  # noqa: TRY301
        except JsonResponseWithError as exc:
            assert exc.args[0] == "Paperless [error]: sample string"  # noqa: PT017

        try:
            payload = {"failure": "something failed"}
            raise JsonResponseWithError(payload)  # noqa: TRY301
        except JsonResponseWithError as exc:
            assert exc.args[0] == "Paperless [failure]: something failed"  # noqa: PT017

        try:
            payload = {"error": ["that", "should", "have", "been", "never", "happened"]}
            raise JsonResponseWithError(payload)  # noqa: TRY301
        except JsonResponseWithError as exc:
            assert exc.args[0] == "Paperless [error]: that"  # noqa: PT017

        try:
            payload = [{"some": [[{"weird": {"error": ["occurred"]}}]]}]
            raise JsonResponseWithError(payload)  # noqa: TRY301
        except JsonResponseWithError as exc:
            assert exc.args[0] == "Paperless [some -> weird -> error]: occurred"  # noqa: PT017

    async def test_request(self, resp: aioresponses) -> None:
        """Test generate request."""
        # we need to use an unmocked PaperlessSession.request() method
        # simply don't initialize Paperless and everything will be fine
        api = Paperless(
            PAPERLESS_TEST_URL,
            PAPERLESS_TEST_TOKEN,
        )

        # test ordinary 200
        resp.get(
            PAPERLESS_TEST_URL,
            status=200,
        )
        async with api.request("get", PAPERLESS_TEST_URL) as res:
            assert res.status

        # last but not least, we test sending a form to test the converter
        form_data = {
            "none_field": None,
            "str_field": "Hello Bytes!",
            "bytes_field": b"Hello String!",
            "tuple_field": (b"Document Content", "filename.pdf"),
            "int_field": 23,
            "float_field": 13.37,
            "int_list": [1, 1, 2, 3, 5, 8, 13],
            "dict_field": {"dict_str_field": "str", "dict_int_field": 2},
        }
        resp.post(
            PAPERLESS_TEST_URL,
            status=200,
        )
        async with api.request("post", PAPERLESS_TEST_URL, form=form_data) as res:
            assert res.status

        # session is still open
        await api.close()

    async def test_request_json(self, resp: aioresponses, api: Paperless) -> None:
        """Test requests."""
        # test 400 bad request with error payload
        resp.get(
            f"{PAPERLESS_TEST_URL}/400-json-error-payload",
            status=400,
            headers={"Content-Type": "application/json"},
            payload={"error": "sample message"},
        )
        with pytest.raises(JsonResponseWithError):
            await api.request_json("get", f"{PAPERLESS_TEST_URL}/400-json-error-payload")

        # test 200 ok with wrong content type
        resp.get(
            f"{PAPERLESS_TEST_URL}/200-text-error-payload",
            status=200,
            headers={"Content-Type": "text/plain"},
            body='{"error": "sample message"}',
        )
        with pytest.raises(BadJsonResponseError):
            await api.request_json("get", f"{PAPERLESS_TEST_URL}/200-text-error-payload")

        # test 200 ok with correct content type, but no json payload
        resp.get(
            f"{PAPERLESS_TEST_URL}/200-json-text-body",
            status=200,
            headers={"Content-Type": "application/json"},
            body="test 5 23 42 1337",
        )
        with pytest.raises(BadJsonResponseError):
            await api.request_json("get", f"{PAPERLESS_TEST_URL}/200-json-text-body")

    async def test_create_url(self) -> None:
        """Test create url util."""
        create_url = Paperless._create_base_url  # pylint: disable=protected-access

        # test default ssl
        url = create_url("hostname")
        assert f"{url.host}" == "hostname"
        assert int(url.port) == 443

        # test enforce http
        url = create_url("http://hostname")
        assert int(url.port) == 80

        # test non-http scheme
        url = create_url("ftp://hostname")
        assert f"{url.scheme}" == "https"

        # should be https even on just setting a port number
        url = create_url("hostname:80")
        assert f"{url.scheme}" == "https"

        # test api/api url
        url = create_url("hostname/api/api/")
        assert f"{url}" == "https://hostname/api/api"

        # test slashes
        url = create_url("hostname/api/endpoint///")
        assert f"{url}" == "https://hostname/api/endpoint"

    async def test_generate_api_token(self, resp: aioresponses, api: Paperless) -> None:
        """Test generate api token."""
        # test successful token creation
        resp.post(
            f"{PAPERLESS_TEST_URL}{API_PATH['token']}",
            status=200,
            payload=DATA_TOKEN,
        )
        token = await api.generate_api_token(
            PAPERLESS_TEST_URL,
            PAPERLESS_TEST_USER,
            PAPERLESS_TEST_PASSWORD,
        )
        assert token == PAPERLESS_TEST_TOKEN

        # test token creation with wrong json answer
        resp.post(
            f"{PAPERLESS_TEST_URL}{API_PATH['token']}",
            status=200,
            payload={"blah": "any string"},
        )
        with pytest.raises(BadJsonResponseError):
            token = await api.generate_api_token(
                PAPERLESS_TEST_URL,
                PAPERLESS_TEST_USER,
                PAPERLESS_TEST_PASSWORD,
            )

        # test error 400
        resp.post(
            f"{PAPERLESS_TEST_URL}{API_PATH['token']}",
            status=400,
            payload={"non_field_errors": ["Unable to log in."]},
        )
        with pytest.raises(JsonResponseWithError):
            token = await api.generate_api_token(
                PAPERLESS_TEST_URL,
                PAPERLESS_TEST_USER,
                PAPERLESS_TEST_PASSWORD,
            )

        # general exception
        resp.post(
            f"{PAPERLESS_TEST_URL}{API_PATH['token']}",
            exception=ValueError,
        )
        with pytest.raises(ValueError):  # noqa: PT011
            token = await api.generate_api_token(
                PAPERLESS_TEST_URL,
                PAPERLESS_TEST_USER,
                PAPERLESS_TEST_PASSWORD,
            )

    async def test_generate_api_token_with_session(
        self, resp: aioresponses, api: Paperless
    ) -> None:
        """Test generate api token with custom session."""
        session = aiohttp.ClientSession()

        resp.post(
            f"{PAPERLESS_TEST_URL}{API_PATH['token']}",
            status=200,
            payload=DATA_TOKEN,
        )
        token = await api.generate_api_token(
            PAPERLESS_TEST_URL,
            PAPERLESS_TEST_USER,
            PAPERLESS_TEST_PASSWORD,
            session=session,
        )
        assert token == PAPERLESS_TEST_TOKEN

    async def test_types(self) -> None:
        """Test types."""
        never_str = "!never_existing_type!"
        never_int = 99952342
        assert PaperlessResource(never_str) == PaperlessResource.UNKNOWN
        assert CustomFieldType(never_str) == CustomFieldType.UNKNOWN
        assert MatchingAlgorithmType(never_int) == MatchingAlgorithmType.UNKNOWN
        assert ShareLinkFileVersionType(never_str) == ShareLinkFileVersionType.UNKNOWN
        assert StatusType(never_str) == StatusType.UNKNOWN
        assert TaskType(never_str) == TaskType.UNKNOWN
        assert TaskStatusType(never_str) == TaskStatusType.UNKNOWN
        assert WorkflowActionType(never_int) == WorkflowActionType.UNKNOWN
        assert WorkflowTriggerType(never_int) == WorkflowTriggerType.UNKNOWN
        assert (
            WorkflowTriggerScheduleDateFieldType(never_str)
            == WorkflowTriggerScheduleDateFieldType.UNKNOWN
        )
        assert WorkflowTriggerSourceType(never_int) == WorkflowTriggerSourceType.UNKNOWN

    async def test_custom_field_draft_value_wo_cache(self, paperless: Paperless) -> None:
        """Test draft custom field value without cache."""
        custom_field = CustomField.create_with_data(
            paperless,
            data={"id": 1337, "name": "Test", "data_type": CustomFieldType.INTEGER},
            fetched=True,
        )
        field_value = custom_field.draft_value(1337)
        for value_type in CUSTOM_FIELD_TYPE_VALUE_MAP.values():
            assert not isinstance(field_value, value_type)

    async def test_custom_field_draft_value_wslash_cache(
        self, resp: aioresponses, paperless: Paperless
    ) -> None:
        """Test draft custom field value with cache."""
        # set custom fields cache
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH['custom_fields']}" + r"\?.*$"),
            status=200,
            payload=DATA_CUSTOM_FIELDS,
        )
        paperless.cache.custom_fields = await paperless.custom_fields.as_dict()

        custom_field = CustomField.create_with_data(
            api=paperless,
            data=DATA_CUSTOM_FIELDS["results"][5],
            fetched=True,
        )
        field_value = custom_field.draft_value(1337, expected_type=CustomFieldIntegerValue)
        assert isinstance(field_value, CustomFieldIntegerValue)

    async def test_custom_field_date_value(self) -> None:
        """Test `CustomFieldDateValue`."""
        test = CustomFieldDateValue(value="1900-01-02")
        assert isinstance(test.value, date)
        test = CustomFieldDateValue(value="1900-01-02T03:04:05.133337Z")
        assert isinstance(test.value, date)

    async def test_custom_field_monetary_value(self) -> None:
        """Test `CustomFieldMonetaryValue`."""
        field = CustomFieldMonetaryValue(value="EUR1337.00")
        assert field.currency == "EUR"
        assert field.amount == 1337

        field.amount = 123.45678
        assert field.amount == 123.46  # round

        field.extra_data = {"default_currency": "USD"}
        assert field.value == "EUR123.46"

        field.value = "123.45"  # no currency
        assert field.currency == "USD"

        field.extra_data = {}
        assert field.currency == ""

        field.currency = "EUR"
        assert field.value == "EUR123.45"

        field.currency = ""
        assert field.value == "123.45"

        field.value = None
        assert field.amount is None

    async def test_custom_field_select_value(self) -> None:
        """Test custom field value types."""
        test = CustomFieldSelectValue(
            value="id2",
            extra_data={
                "select_options": [
                    {"id": "id1", "label": "label1"},
                    {"id": "id2", "label": "label2"},
                ]
            },
        )
        assert isinstance(test.labels, list)
        assert test.label == "label2"

        # test fail
        test.extra_data = None
        assert test.label is None

    async def test_dataclass_conversion(self) -> None:  # pylint: disable=too-many-statements
        """Test dataclass utils."""

        class SomeStatus(Enum):
            """Test enum."""

            ACTIVE = 1
            INACTIVE = 2
            UNKNOWN = -1

            @classmethod
            def _missing_(cls: "SomeStatus", *_: object) -> "SomeStatus":
                """Set default."""
                return cls.UNKNOWN

        class SomeNestedExtraData(TypedDict):
            """Test nested TypedDict."""

            ustr: str | None
            uany: int | str | bool | None

        class SomeExtraData(TypedDict):
            """Test TypedDict."""

            a_str: str
            a_dict: dict[str, str]
            a_list: list[str]
            a_typeddict: SomeNestedExtraData

        @dataclass
        class SomeFriend:
            """Test class."""

            name: str
            age: int

            @classmethod
            def from_dict(cls, data: dict) -> "SomeFriend":
                """Test from_dict stuff."""
                return cls(name=str(data.get("name")), age=int(data.get("age")))

        @dataclass
        class SomePerson:
            """Test class."""

            name: str
            age: int
            height: float
            height2: float
            birth: date
            last_login: datetime
            friends: list[SomeFriend] | None
            deleted: datetime | None
            is_deleted: bool
            status: SomeStatus
            file: bytes
            meta: dict[str, str]
            extra_data: SomeExtraData

        raw_data = {
            "name": "Lee Tobi, Sajangnim",
            "age": 38,
            "height": 1.76,
            "height2": 2,
            "birth": "1986-05-23",
            "last_login": "2023-08-08T06:06:35.495972Z",
            "is_deleted": False,
            "friends": [
                {
                    "name": "Erika",
                    "age": "50",  # this should be int, check "back conversion" at bottom
                },
                {
                    "name": "Reinhard",
                    "age": 40,
                },
            ],
            "status": 1,
            "file": b"5-23-42-666-0815-1337",
            "meta": {"hairs": "blonde", "eyes": "blue", "loves": "Python"},
            "extra_data": {
                "a_str": "test",
                "a_dict": {
                    "key1": "val1",
                    "key2": "val2",
                },
                "a_list": ["a", "b", "c"],
                "a_typeddict": {"ustr": "hello", "uany": 1},
            },
        }

        data = {
            field.name: dict_value_to_object(
                f"_Person.{__name__}.{field.name}",
                raw_data.get(field.name),
                field.type,
                field.default,
            )
            for field in fields(SomePerson)
        }
        res = SomePerson(**data)

        assert isinstance(res.name, str)
        assert isinstance(res.age, int)
        assert isinstance(res.height, float)
        assert isinstance(res.height2, float)
        assert isinstance(res.birth, date)
        assert isinstance(res.last_login, datetime)
        assert isinstance(res.friends, list)
        assert isinstance(res.friends[0], SomeFriend)
        assert isinstance(res.friends[0].age, int)
        assert isinstance(res.friends[1].age, int)
        assert res.deleted is None
        assert res.is_deleted is False
        assert isinstance(res.status, SomeStatus)
        assert isinstance(res.file, bytes)
        assert isinstance(res.extra_data, dict)
        assert isinstance(res.extra_data["a_typeddict"], dict)

        # back conversion
        back = {field.name: object_to_dict_value(getattr(res, field.name)) for field in fields(res)}

        assert isinstance(back["friends"][0]["age"], int)  # was str in the source dict
        assert isinstance(back["meta"], dict)
        assert isinstance(back["extra_data"], dict)
        assert isinstance(back["extra_data"]["a_list"], list)

    async def test_pages_object(self, api: Paperless) -> None:
        """Test pages."""

        @dataclass(init=False)
        class TestResource(PaperlessModel):
            """Test Resource."""

            id: int | None = None

        data = {
            "count": 0,
            "current_page": 1,
            "page_size": 25,
            "next": "any.url",
            "previous": None,
            "all": [],
            "results": [],
        }

        for i in range(1, 101):
            data["count"] += 1
            data["all"].append(i)
            data["results"].append({"id": i})

        page = Page.create_with_data(api, data=data, fetched=True)
        page._resource_cls = TestResource  # pylint: disable=protected-access

        assert isinstance(page, Page)
        assert page.current_count == 100
        for item in page:
            assert isinstance(item, TestResource)

        # check first page
        assert not page.has_previous_page
        assert page.has_next_page
        assert not page.is_last_page
        assert page.last_page == 4
        assert page.next_page == 2
        assert page.previous_page is None

        # check inner page
        page.previous = "any.url"
        page.current_page = 3

        assert page.previous_page is not None
        assert page.next_page is not None
        assert not page.is_last_page

        # check last page
        page.next = None
        page.current_page = 4

        assert page.next_page is None
        assert page.is_last_page

    async def test_draft_exc(self, api: Paperless) -> None:
        """Test draft not supported."""

        @dataclass(init=False)
        class TestResource(PaperlessModel):
            """Test Resource."""

        class TestHelper(HelperBase, helpers.DraftableMixin):
            """Test Helper."""

            _api_path = "any.url"
            _resource = "test"
            # draft_cls - we "forgot" to set a draft class, which will raise
            _resource_cls = TestResource

        helper = TestHelper(api)
        with pytest.raises(DraftNotSupportedError):
            # ... there it is
            helper.draft()



"""Tags snapshot."""

DATA_TAGS = {
    "count": 2,
    "next": None,
    "previous": None,
    "all": [
        1,
        2,
        3,
        4,
        5,
    ],
    "results": [
        {
            "id": 1,
            "slug": "important",
            "name": "Important",
            "color": "#ff0000",
            "text_color": "#00ff00",
            "match": "",
            "matching_algorithm": 0,
            "is_insensitive": True,
            "is_inbox_tag": False,
            "document_count": 20,
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 2,
            "slug": "inbox",
            "name": "Inbox",
            "color": "#ff0000",
            "text_color": "#00ff00",
            "match": "",
            "matching_algorithm": 1,
            "is_insensitive": True,
            "is_inbox_tag": True,
            "document_count": 20,
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 3,
            "slug": "test-3",
            "name": "Test 3",
            "color": "#ff0000",
            "text_color": "#00ff00",
            "match": "",
            "matching_algorithm": 2,
            "is_insensitive": True,
            "is_inbox_tag": False,
            "document_count": 20,
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 4,
            "slug": "test-4",
            "name": "Test 4",
            "color": "#ff0000",
            "text_color": "#00ff00",
            "match": "",
            "matching_algorithm": 3,
            "is_insensitive": True,
            "is_inbox_tag": False,
            "document_count": 20,
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 5,
            "slug": "test-5",
            "name": "Test 5",
            "color": "#ff0000",
            "text_color": "#00ff00",
            "match": "",
            "matching_algorithm": 4,
            "is_insensitive": True,
            "is_inbox_tag": False,
            "document_count": 20,
            "owner": None,
            "user_can_change": True,
        },
    ],
}



"""PyPaperless exceptions."""

from typing import Any


class PaperlessError(Exception):
    """Base exception for PyPaperless."""


# Sessions and requests


class InitializationError(PaperlessError):
    """Raise when initializing a `Paperless` instance without valid url or token."""


class PaperlessConnectionError(InitializationError, PaperlessError):
    """Raise when connection to Paperless is not possible."""


class PaperlessAuthError(InitializationError, PaperlessError):
    """Raise when response is 401 code."""


class PaperlessInvalidTokenError(PaperlessAuthError):
    """Raise when response is 401 due invalid access token."""


class PaperlessInactiveOrDeletedError(PaperlessAuthError):
    """Raise when response is 401 code due user is inactive or deleted."""


class PaperlessForbiddenError(InitializationError, PaperlessError):
    """Raise when response is 403 code."""


class BadJsonResponseError(PaperlessError):
    """Raise when response is no valid json."""


class JsonResponseWithError(PaperlessError):
    """Raise when Paperless accepted the request, but responded with an error payload."""

    def __init__(self, payload: Any) -> None:
        """Initialize a `JsonResponseWithError` instance."""

        def _parse_payload(payload: Any, key: list[str] | None = None) -> tuple[list[str], str]:
            """Parse first suitable error from payload."""
            if key is None:
                key = []

            if isinstance(payload, list):
                return _parse_payload(payload.pop(0), key)
            if isinstance(payload, dict):
                if "error" in payload:
                    key.append("error")
                    return _parse_payload(payload["error"], key)

                new_key = next(iter(payload))
                key.append(new_key)

                return _parse_payload(payload[new_key], key)

            return key, payload

        key, message = _parse_payload(payload)

        if len(key) == 0:
            key.append("error")
        key_chain = " -> ".join(key)

        super().__init__(f"Paperless [{key_chain}]: {message}")


# Models


class AsnRequestError(PaperlessError):
    """Raise when getting an error during requesting the next asn."""


class DraftFieldRequiredError(PaperlessError):
    """Raise when trying to save models with missing required fields."""


class DraftNotSupportedError(PaperlessError):
    """Raise when trying to draft unsupported models."""


class ItemNotFoundError(PaperlessError):
    """Raise when trying to access non-existing items in PaperlessModelData classes."""


class PrimaryKeyRequiredError(PaperlessError):
    """Raise when trying to access model data without supplying a pk."""


class SendEmailError(PaperlessError):
    """Raise when sending email for a document fails."""


# Tasks


class TaskNotFoundError(PaperlessError):
    """Raise when trying to access a task by non-existing uuid."""

    def __init__(self, task_id: str) -> None:
        """Initialize a `TaskNotFound` instance."""
        super().__init__(f"Task with UUID {task_id} not found.")



"""Users snapshot."""

DATA_USERS = {
    "count": 4,
    "next": None,
    "previous": None,
    "all": [1, 2, 3, 4],
    "results": [
        {
            "id": 1,
            "username": "admin",
            "email": "you@inside.me",
            "password": "**********",
            "first_name": "",
            "last_name": "",
            "date_joined": "2022-02-11T22:28:25+00:00",
            "is_staff": True,
            "is_active": False,
            "is_superuser": False,
            "groups": [],
            "user_permissions": [],
            "inherited_permissions": [],
        },
        {
            "id": 2,
            "username": "alpha",
            "email": "alpha@is-not.beta",
            "password": "**********",
            "first_name": "Alpha",
            "last_name": "Centauri",
            "date_joined": "2023-06-27T15:59:01.975496+00:00",
            "is_staff": False,
            "is_active": True,
            "is_superuser": False,
            "groups": [1],
            "user_permissions": [],
            "inherited_permissions": [
                "documents.delete_customfieldinstance",
                "documents.delete_consumptiontemplate",
                "documents.delete_customfield",
                "documents.add_uisettings",
                "documents.view_uisettings",
                "documents.view_tag",
                "documents.change_customfield",
                "documents.view_sharelink",
                "documents.change_note",
                "documents.view_correspondent",
                "documents.change_customfieldinstance",
                "documents.delete_note",
                "documents.add_consumptiontemplate",
                "documents.add_customfield",
                "documents.change_sharelink",
                "documents.view_consumptiontemplate",
                "documents.view_documenttype",
                "documents.change_document",
                "documents.view_customfieldinstance",
                "documents.add_sharelink",
                "documents.view_document",
                "documents.delete_uisettings",
                "documents.add_savedview",
                "documents.view_note",
                "documents.view_customfield",
                "documents.add_customfieldinstance",
                "documents.change_uisettings",
                "documents.change_savedview",
                "documents.add_document",
                "documents.add_note",
                "documents.change_consumptiontemplate",
                "documents.delete_sharelink",
                "documents.delete_savedview",
                "documents.view_savedview",
                "documents.delete_document",
            ],
        },
        {
            "id": 3,
            "username": "omega",
            "email": "omega.weapon@ff8.net",
            "password": "**********",
            "first_name": "Omega",
            "last_name": "Weapon",
            "date_joined": "2022-02-14T13:42:54+00:00",
            "is_staff": True,
            "is_active": True,
            "is_superuser": True,
            "groups": [1],
            "user_permissions": [],
            "inherited_permissions": [
                "documents.change_paperlesstask",
                "documents.delete_customfieldinstance",
                "django_celery_results.delete_chordcounter",
                "paperless_mail.change_processedmail",
                "django_celery_results.view_chordcounter",
                "documents.view_storagepath",
                "documents.delete_tag",
                "sessions.delete_session",
                "documents.view_log",
                "django_celery_results.change_groupresult",
                "documents.add_uisettings",
                "authtoken.delete_token",
                "admin.add_logentry",
                "django_celery_results.add_groupresult",
                "guardian.view_userobjectpermission",
                "paperless_mail.add_processedmail",
                "documents.change_customfieldinstance",
                "documents.view_correspondent",
                "auth.add_permission",
                "admin.view_logentry",
                "documents.delete_note",
                "documents.add_customfield",
                "auth.view_permission",
                "documents.add_consumptiontemplate",
                "guardian.delete_groupobjectpermission",
                "documents.change_sharelink",
                "documents.view_consumptiontemplate",
                "documents.delete_log",
                "guardian.delete_userobjectpermission",
                "paperless_mail.delete_mailrule",
                "sessions.change_session",
                "documents.add_tag",
                "documents.view_documenttype",
                "auth.view_user",
                "auth.add_group",
                "guardian.view_groupobjectpermission",
                "documents.change_document",
                "documents.change_savedviewfilterrule",
                "paperless_mail.add_mailrule",
                "documents.view_customfieldinstance",
                "guardian.add_userobjectpermission",
                "documents.add_sharelink",
                "auth.change_user",
                "documents.view_document",
                "django_celery_results.delete_taskresult",
                "documents.delete_uisettings",
                "documents.add_savedview",
                "documents.view_note",
                "authtoken.view_token",
                "authtoken.change_tokenproxy",
                "documents.change_storagepath",
                "documents.change_uisettings",
                "django_celery_results.add_chordcounter",
                "paperless_mail.change_mailrule",
                "documents.change_documenttype",
                "documents.change_savedview",
                "auth.delete_permission",
                "documents.add_storagepath",
                "django_celery_results.change_chordcounter",
                "documents.add_note",
                "documents.change_consumptiontemplate",
                "documents.delete_savedview",
                "guardian.change_userobjectpermission",
                "admin.delete_logentry",
                "authtoken.add_token",
                "documents.view_savedview",
                "documents.delete_paperlesstask",
                "sessions.view_session",
                "auth.delete_group",
                "paperless_mail.delete_mailaccount",
                "documents.add_savedviewfilterrule",
                "authtoken.delete_tokenproxy",
                "documents.delete_consumptiontemplate",
                "authtoken.view_tokenproxy",
                "authtoken.change_token",
                "documents.delete_customfield",
                "contenttypes.change_contenttype",
                "documents.change_log",
                "paperless_mail.view_mailrule",
                "documents.view_uisettings",
                "documents.view_tag",
                "documents.change_customfield",
                "admin.change_logentry",
                "documents.view_sharelink",
                "documents.change_note",
                "documents.delete_savedviewfilterrule",
                "auth.add_user",
                "django_celery_results.delete_groupresult",
                "documents.change_correspondent",
                "documents.add_documenttype",
                "django_celery_results.view_groupresult",
                "documents.add_paperlesstask",
                "contenttypes.add_contenttype",
                "auth.view_group",
                "guardian.change_groupobjectpermission",
                "django_celery_results.change_taskresult",
                "documents.change_tag",
                "paperless_mail.delete_processedmail",
                "paperless_mail.change_mailaccount",
                "documents.delete_correspondent",
                "paperless_mail.view_processedmail",
                "documents.delete_storagepath",
                "sessions.add_session",
                "paperless_mail.add_mailaccount",
                "django_celery_results.view_taskresult",
                "documents.view_customfield",
                "documents.view_paperlesstask",
                "documents.add_customfieldinstance",
                "django_celery_results.add_taskresult",
                "documents.view_savedviewfilterrule",
                "documents.add_log",
                "guardian.add_groupobjectpermission",
                "auth.delete_user",
                "auth.change_permission",
                "documents.add_document",
                "documents.delete_documenttype",
                "documents.delete_sharelink",
                "documents.add_correspondent",
                "authtoken.add_tokenproxy",
                "auth.change_group",
                "contenttypes.delete_contenttype",
                "contenttypes.view_contenttype",
                "paperless_mail.view_mailaccount",
                "documents.delete_document",
            ],
        },
        {
            "id": 4,
            "username": "zg_test",
            "email": "",
            "password": "**********",
            "first_name": "",
            "last_name": "",
            "date_joined": "2023-06-27T15:56:39.754859+00:00",
            "is_staff": False,
            "is_active": True,
            "is_superuser": False,
            "groups": [1],
            "user_permissions": [],
            "inherited_permissions": [
                "documents.delete_customfieldinstance",
                "documents.delete_consumptiontemplate",
                "documents.delete_customfield",
                "documents.add_uisettings",
                "documents.view_uisettings",
                "documents.view_tag",
                "documents.change_customfield",
                "documents.view_sharelink",
                "documents.change_note",
                "documents.view_correspondent",
                "documents.change_customfieldinstance",
                "documents.delete_note",
                "documents.add_consumptiontemplate",
                "documents.add_customfield",
                "documents.change_sharelink",
                "documents.view_consumptiontemplate",
                "documents.view_documenttype",
                "documents.change_document",
                "documents.view_customfieldinstance",
                "documents.add_sharelink",
                "documents.view_document",
                "documents.delete_uisettings",
                "documents.add_savedview",
                "documents.view_note",
                "documents.view_customfield",
                "documents.add_customfieldinstance",
                "documents.change_uisettings",
                "documents.change_savedview",
                "documents.add_document",
                "documents.add_note",
                "documents.change_consumptiontemplate",
                "documents.delete_sharelink",
                "documents.delete_savedview",
                "documents.view_savedview",
                "documents.delete_document",
            ],
        },
    ],
}



"""Provide `Remote Version` related models and helpers."""

from dataclasses import dataclass

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel


@dataclass(init=False)
class RemoteVersion(PaperlessModel):
    """Represent Paperless `Remote Version`."""

    _api_path = API_PATH["remote_version"]

    version: str | None = None
    update_available: bool | None = None


class RemoteVersionHelper(HelperBase):
    """Represent a factory for Paperless `Remote Version` models."""

    _api_path = API_PATH["remote_version"]
    _resource = PaperlessResource.REMOTE_VERSION

    _resource_cls = RemoteVersion

    async def __call__(self) -> RemoteVersion:
        """Request the `Remote Version` model data."""
        res = await self._api.request_json("get", self._api_path)
        return self._resource_cls.create_with_data(self._api, res, fetched=True)



"""Status snapshot."""

DATA_STATUS = {
    "pngx_version": "2.6.1",
    "server_os": "Linux-6.6.12-linuxkit-aarch64-with-glibc2.36",
    "install_type": "docker",
    "storage": {
        "total": 494384795648,
        "available": 103324229632,
    },
    "database": {
        "type": "sqlite",
        "url": "/usr/src/paperless/data/db.sqlite3",
        "status": "OK",
        "error": None,
        "migration_status": {
            "latest_migration": "paperless.0003_alter_applicationconfiguration_max_image_pixels",
            "unapplied_migrations": [],
        },
    },
    "tasks": {
        "redis_url": "redis://broker:6379",
        "redis_status": "OK",
        "redis_error": None,
        "celery_status": "OK",
        "index_status": "OK",
        "index_last_modified": "2024-03-06T07:10:55.370884+01:00",
        "index_error": None,
        "classifier_status": "OK",
        "classifier_last_trained": "2024-03-06T07:05:01.281804+00:00",
        "classifier_error": None,
    },
}



"""Share links snapshot."""

DATA_SHARE_LINKS = {
    "count": 5,
    "next": None,
    "previous": None,
    "all": [1, 2, 3, 4, 5, 6, 7, 8],
    "results": [
        {
            "id": 1,
            "created": "2023-12-11T14:06:49.096456Z",
            "expiration": "2023-12-18T14:06:49.064000Z",
            "slug": "GMIFR9WVPe7a0FAltmrAdmVsrrTzH6Z9yFi2jufhi5yCTAMWfF",
            "document": 1,
            "file_version": "original",
        },
        {
            "id": 2,
            "created": "2023-12-11T14:06:53.583496Z",
            "expiration": "2024-01-10T14:06:53.558000Z",
            "slug": "Px2h3mrkIvExyTE8M8usrTLv3jtTb4MnLJ4eTAxcjy2FUmuDLq",
            "document": 2,
            "file_version": "original",
        },
        {
            "id": 3,
            "created": "2023-12-11T14:06:55.984583Z",
            "expiration": None,
            "slug": "bDnxeQ4UmlFVUYCDrb1KBLbE4HVSW8jw3CLElcwPyAncV5eiI+00:00",
            "document": 1,
            "file_version": "original",
        },
        {
            "id": 4,
            "created": "2023-12-11T14:07:01.448813Z",
            "expiration": "2023-12-12T14:07:01.423000Z",
            "slug": "HfzHhDzA03ZQg4t4TAlOuup59qgQA18Zjbb9eOE06PZ8KTjgOb",
            "document": 2,
            "file_version": "archive",
        },
        {
            "id": 5,
            "created": "2023-12-11T14:11:50.710369Z",
            "expiration": None,
            "slug": "7PIGEZbeFv5yIrnpSVwj1QeXiJu0IZCiEWGIV4aUHQrfUQtXne",
            "document": 1,
            "file_version": "archive",
        },
        {
            "id": 6,
            "created": "2023-12-11T14:11:50.710369Z",
            "expiration": None,
            "slug": "7PIGEZbeFv5yIrnpSVwj1QeXiJu0IZCiEWGIV4aUHQrfUQtXne",
            "document": 1,
            "file_version": "archive",
        },
        {
            "id": 7,
            "created": "2023-12-11T14:11:50.710369Z",
            "expiration": None,
            "slug": "7PIGEZbeFv5yIrnpSVwj1QeXiJu0IZCiEWGIV4aUHQrfUQtXne",
            "document": 1,
            "file_version": "archive",
        },
        {
            "id": 8,
            "created": "2023-12-11T14:11:50.710369Z",
            "expiration": None,
            "slug": "7PIGEZbeFv5yIrnpSVwj1QeXiJu0IZCiEWGIV4aUHQrfUQtXne",
            "document": 1,
            "file_version": "archive",
        },
    ],
}



"""Provide `Statistics` related models and helpers."""

from dataclasses import dataclass

from pypaperless.const import API_PATH, PaperlessResource

from .base import HelperBase, PaperlessModel
from .common import StatisticDocumentFileTypeCount


@dataclass(init=False)
class Statistic(PaperlessModel):
    """Represent Paperless `Statistic`."""

    _api_path = API_PATH["statistics"]

    documents_total: int | None = None
    documents_inbox: int | None = None
    inbox_tag: int | None = None
    inbox_tags: list[int] | None = None
    document_file_type_counts: list[StatisticDocumentFileTypeCount] | None = None
    character_count: int | None = None
    tag_count: int | None = None
    correspondent_count: int | None = None
    document_type_count: int | None = None
    storage_path_count: int | None = None
    current_asn: int | None = None


class StatisticHelper(HelperBase):
    """Represent a factory for Paperless `Statistic` models."""

    _api_path = API_PATH["statistics"]
    _resource = PaperlessResource.STATISTICS

    _resource_cls = Statistic

    async def __call__(self) -> Statistic:
        """Request the `Statistic` model data."""
        res = await self._api.request_json("get", self._api_path)
        return self._resource_cls.create_with_data(self._api, res, fetched=True)



"""Paths snapshot."""

from tests.const import PAPERLESS_TEST_URL

DATA_PATHS = {
    "correspondents": f"{PAPERLESS_TEST_URL}/api/correspondents/",
    "document_types": f"{PAPERLESS_TEST_URL}/api/document_types/",
    "documents": f"{PAPERLESS_TEST_URL}/api/documents/",
    "logs": f"{PAPERLESS_TEST_URL}/api/logs/",
    "tags": f"{PAPERLESS_TEST_URL}/api/tags/",
    "saved_views": f"{PAPERLESS_TEST_URL}/api/saved_views/",
    "tasks": f"{PAPERLESS_TEST_URL}/api/tasks/",
    "users": f"{PAPERLESS_TEST_URL}/api/users/",
    "groups": f"{PAPERLESS_TEST_URL}/api/groups/",
    "mail_accounts": f"{PAPERLESS_TEST_URL}/api/mail_accounts/",
    "mail_rules": f"{PAPERLESS_TEST_URL}/api/mail_rules/",
    "storage_paths": f"{PAPERLESS_TEST_URL}/api/storage_paths/",
    "config": f"{PAPERLESS_TEST_URL}/api/config/",
    "custom_fields": f"{PAPERLESS_TEST_URL}/api/custom_fields/",
    "share_links": f"{PAPERLESS_TEST_URL}/api/share_links/",
    "workflows": f"{PAPERLESS_TEST_URL}/api/workflows/",
    "workflow_actions": f"{PAPERLESS_TEST_URL}/api/workflow_actions/",
    "workflow_triggers": f"{PAPERLESS_TEST_URL}/api/workflow_triggers/",
}



"""Provide `Document` related models and helpers."""

import datetime
from collections.abc import AsyncGenerator, Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Self, cast, overload

from pypaperless.const import API_PATH, PaperlessResource
from pypaperless.exceptions import (
    AsnRequestError,
    ItemNotFoundError,
    PrimaryKeyRequiredError,
    SendEmailError,
)
from pypaperless.models.utils import object_to_dict_value

from .base import HelperBase, PaperlessModel, PaperlessModelData
from .common import (
    CUSTOM_FIELD_TYPE_VALUE_MAP,
    CustomFieldType,
    CustomFieldValue,
    CustomFieldValueT,
    DocumentMetadataType,
    DocumentSearchHitType,
    RetrieveFileMode,
)
from .custom_fields import CustomField
from .mixins import helpers, models

if TYPE_CHECKING:
    from pypaperless import Paperless


class DocumentCustomFieldList(PaperlessModelData):
    """Represent a list of Paperless custom field instances typically on documents."""

    def __init__(self, api: "Paperless", data: list[dict[str, Any]]) -> None:
        """Initialize a `DocumentCustomFieldList` instance."""
        self._api = api
        self._data = data
        self._fields: list[CustomFieldValue] = []

        cache = api.cache.custom_fields

        for item in data:
            if cache and (field := cache.get(item["field"], None)):
                klass = CUSTOM_FIELD_TYPE_VALUE_MAP.get(
                    field.data_type or CustomFieldType.UNKNOWN, CustomFieldValue
                )
                klass_data = {
                    **item,
                    "name": field.name,
                    "data_type": field.data_type,
                    "extra_data": field.extra_data,
                }
                self._fields.append(klass(**klass_data))
            else:
                self._fields.append(CustomFieldValue(**item))

    def __contains__(self, field: int | CustomField) -> bool:
        """Check if the given `CustomField` or its id is present in `DocumentCustomFieldList`."""
        item_id = field.id if isinstance(field, CustomField) else field
        return any(item.field == item_id for item in self._fields)

    def __iter__(self) -> Iterator[CustomFieldValue]:
        """Iterate over custom fields.

        Example:
        -------
        ```python
        for item in document.custom_fields:
            # do something
        ```

        """
        yield from self._fields

    def __iadd__(self, field: CustomFieldValue) -> Self:
        """Add a new `CustomFieldValue` to a document."""
        return self.add(field)

    def add(self, field: CustomFieldValue) -> Self:
        """Add a new `CustomFieldValue` to a document."""
        self._fields.append(field)
        return self

    def __isub__(self, field: CustomFieldValue | CustomField | int) -> Self:
        """Remove a `CustomFieldValue` from a document."""
        return self.remove(field)

    def remove(self, field: CustomFieldValue | CustomField | int) -> Self:
        """Remove a `CustomFieldValue` from a document."""
        item_id = (
            field.id
            if isinstance(field, CustomField)
            else field.field
            if isinstance(field, CustomFieldValue)
            else field
        )
        self._fields = [field for field in self._fields if field.field != item_id]

        return self

    @overload
    def default(self, field: int | CustomField) -> CustomFieldValue | None: ...

    @overload
    def default(
        self, field: int | CustomField, expected_type: type[CustomFieldValueT]
    ) -> CustomFieldValueT | None: ...

    def default(
        self, field: int | CustomField, expected_type: type[CustomFieldValueT] | None = None
    ) -> CustomFieldValue | CustomFieldValueT | None:
        """Access and return a (typed) `CustomFieldValue`, or `None`."""
        try:
            value = self.get(field)
        except ItemNotFoundError:
            return None

        if expected_type is not None and not isinstance(value, expected_type):
            msg = f"Expected {expected_type.__name__}, got {type(value).__name__}"
            raise TypeError(msg)

        return value

    @overload
    def get(self, field: int | CustomField) -> CustomFieldValue: ...

    @overload
    def get(
        self, field: int | CustomField, expected_type: type[CustomFieldValueT]
    ) -> CustomFieldValueT: ...

    def get(
        self, field: int | CustomField, expected_type: type[CustomFieldValueT] | None = None
    ) -> CustomFieldValue | CustomFieldValueT:
        """Access and return a (typed) `CustomFieldValue` from the list."""
        item_id = field.id if isinstance(field, CustomField) else field

        for item in self._fields:
            if item.field == item_id:
                if expected_type is not None and not isinstance(item, expected_type):
                    msg = f"Expected {expected_type.__name__}, got {type(item).__name__}"
                    raise TypeError(msg)
                return item

        raise ItemNotFoundError

    @classmethod
    def unserialize(cls, api: "Paperless", data: list[dict[str, Any]]) -> Self:
        """Return a new instance of `cls` from `data`.

        Primarily used by `dict_value_to_object` when instantiating model classes.
        """
        return cls(api, data=data)

    def serialize(self) -> list[dict[str, Any]]:
        """Serialize the class data."""
        return [{"field": field.field, "value": field.value} for field in self._fields]


@dataclass(init=False)
class Document(
    PaperlessModel,
    models.SecurableMixin,
    models.UpdatableMixin,
    models.DeletableMixin,
):
    """Represent a Paperless `Document`."""

    _api_path = API_PATH["documents_single"]

    id: int | None = None
    correspondent: int | None = None
    document_type: int | None = None
    storage_path: int | None = None
    title: str | None = None
    content: str | None = None
    tags: list[int] | None = None
    created: datetime.date | None = None
    modified: datetime.datetime | None = None
    added: datetime.datetime | None = None
    deleted_at: datetime.datetime | None = None
    archive_serial_number: int | None = None
    original_file_name: str | None = None
    archived_file_name: str | None = None
    is_shared_by_requester: bool | None = None
    custom_fields: DocumentCustomFieldList | None = None
    page_count: int | None = None
    mime_type: str | None = None
    __search_hit__: DocumentSearchHitType | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `Document` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))
        self.notes = DocumentNoteHelper(api, data.get("id"))

    @property
    def created_date(self) -> datetime.date | None:
        """Backward compatibility for the removed `created_date` field."""
        return self.created

    @property
    def has_search_hit(self) -> bool:
        """Return if the document has a search hit attached."""
        return self.__search_hit__ is not None

    @property
    def search_hit(self) -> DocumentSearchHitType | None:
        """Return the document search hit."""
        return self.__search_hit__

    async def get_download(self, *, original: bool = False) -> "DownloadedDocument":
        """Request and return the `DownloadedDocument` class."""
        return await self._api.documents.download(cast("int", self.id), original=original)

    async def get_metadata(self) -> "DocumentMeta":
        """Request and return the documents `DocumentMeta` class."""
        return await self._api.documents.metadata(cast("int", self.id))

    async def get_preview(self, *, original: bool = False) -> "DownloadedDocument":
        """Request and return the `DownloadedDocument` class."""
        return await self._api.documents.preview(cast("int", self.id), original=original)

    async def get_suggestions(self) -> "DocumentSuggestions":
        """Request and return the `DocumentSuggestions` class."""
        return await self._api.documents.suggestions(cast("int", self.id))

    async def get_thumbnail(self, *, original: bool = False) -> "DownloadedDocument":
        """Request and return the `DownloadedDocument` class."""
        return await self._api.documents.thumbnail(cast("int", self.id), original=original)


@dataclass(init=False)
class DocumentDraft(PaperlessModel, models.CreatableMixin):
    """Represent a new Paperless `Document`, which is not stored in Paperless."""

    _api_path = API_PATH["documents_post"]

    _create_required_fields = {"document"}

    document: bytes | None = None
    filename: str | None = None
    title: str | None = None
    created: datetime.datetime | None = None
    correspondent: int | None = None
    document_type: int | None = None
    storage_path: int | None = None
    tags: int | list[int] | None = None
    archive_serial_number: int | None = None
    custom_fields: list[int] | None = None

    def _serialize(self) -> dict[str, Any]:
        """Serialize."""
        data = {
            "form": {
                field.name: object_to_dict_value(getattr(self, field.name))
                for field in self._get_dataclass_fields()
                if field.name not in {"document", "filename"}
            }
        }
        data["form"].update(
            {
                "document": (
                    (self.document, self.filename) if self.filename is not None else self.document
                )
            }
        )
        return data


@dataclass(init=False)
class DocumentNote(PaperlessModel):
    """Represent a Paperless `DocumentNote`."""

    _api_path = API_PATH["documents_notes"]

    id: int | None = None
    note: str | None = None
    created: datetime.datetime | None = None
    document: int | None = None
    user: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `DocumentNote` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("document"))

    async def delete(self) -> bool:
        """Delete a `resource item` from DRF. There is no point of return.

        Return `True` when deletion was successful, `False` otherwise.

        Example:
        -------
        ```python
        # request document notes
        notes = await paperless.documents.notes(42)

        for note in notes:
            if await note.delete():
                print("Successfully deleted the note!")
        ```

        """
        params = {
            "id": self.id,
        }
        async with self._api.request("delete", self._api_path, params=params) as res:
            return res.status in {200, 204}  # backward compatibility


@dataclass(kw_only=True)
class DocumentNoteDraft(PaperlessModel, models.CreatableMixin):
    """Represent a new Paperless `DocumentNote`, which is not stored in Paperless."""

    _api_path = API_PATH["documents_notes"]

    _create_required_fields = {"note", "document"}

    note: str | None = None
    document: int | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `DocumentNote` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("document"))


@dataclass(init=False)
class DocumentMeta(PaperlessModel):
    """Represent a Paperless `Document`s metadata."""

    _api_path = API_PATH["documents_meta"]

    id: int | None = None
    original_checksum: str | None = None
    original_size: int | None = None
    original_mime_type: str | None = None
    media_filename: str | None = None
    has_archive_version: bool | None = None
    original_metadata: list[DocumentMetadataType] | None = None
    archive_checksum: str | None = None
    archive_media_filename: str | None = None
    original_filename: str | None = None
    lang: str | None = None
    archive_size: int | None = None
    archive_metadata: list[DocumentMetadataType] | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `DocumentMeta` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


@dataclass(init=False)
class DownloadedDocument(PaperlessModel):
    """Represent a Paperless `Document`s downloaded file."""

    _api_path = API_PATH["documents"]

    id: int | None = None
    mode: RetrieveFileMode | None = None
    original: bool | None = None
    content: bytes | None = None
    content_type: str | None = None
    disposition_filename: str | None = None
    disposition_type: str | None = None

    async def load(self) -> None:
        """Get `raw data` from DRF."""
        self._api_path = self._api_path.format(pk=self._data.get("id"))

        params = {
            "original": "true" if self._data.get("original", False) else "false",
        }

        async with self._api.request("get", self._api_path, params=params) as res:
            self._data.update(
                {
                    "content": await res.read(),
                    "content_type": res.content_type,
                }
            )

            if res.content_disposition is not None:
                self._data.update(
                    {
                        "disposition_filename": res.content_disposition.filename,
                        "disposition_type": res.content_disposition.type,
                    }
                )

        self._set_dataclass_fields()
        self._fetched = True


@dataclass(init=False)
class DocumentSuggestions(PaperlessModel):
    """Represent a Paperless `Document` suggestions."""

    _api_path = API_PATH["documents_suggestions"]

    id: int | None = None
    correspondents: list[int] | None = None
    tags: list[int] | None = None
    document_types: list[int] | None = None
    storage_paths: list[int] | None = None
    dates: list[datetime.date] | None = None

    def __init__(self, api: "Paperless", data: dict[str, Any]) -> None:
        """Initialize a `DocumentSuggestions` instance."""
        super().__init__(api, data)

        self._api_path = self._api_path.format(pk=data.get("id"))


class DocumentSuggestionsHelper(HelperBase):
    """Represent a factory for Paperless `DocumentSuggestions` models."""

    _api_path = API_PATH["documents_suggestions"]
    _resource = PaperlessResource.DOCUMENTS

    _resource_cls = DocumentSuggestions

    async def __call__(self, pk: int) -> DocumentSuggestions:
        """Request exactly one resource item."""
        data = {
            "id": pk,
        }
        item = self._resource_cls.create_with_data(self._api, data)
        await item.load()

        return item


class DocumentSubHelperBase(HelperBase):
    """Represent a factory for Paperless `DownloadedDocument` models."""

    _api_path = API_PATH["documents_suggestions"]
    _resource = PaperlessResource.DOCUMENTS

    _resource_cls = DownloadedDocument

    async def __call__(
        self,
        pk: int,
        mode: RetrieveFileMode,
        api_path: str,
        *,
        original: bool,
    ) -> DownloadedDocument:
        """Request exactly one resource item."""
        data = {
            "id": pk,
            "mode": mode,
            "original": original,
        }
        item = self._resource_cls.create_with_data(self._api, data)
        item._api_path = api_path  # noqa: SLF001
        await item.load()

        return item


class DocumentFileDownloadHelper(DocumentSubHelperBase):
    """Represent a factory for Paperless `DownloadedDocument` models."""

    _api_path = API_PATH["documents_download"]

    async def __call__(  # type: ignore[override]
        self,
        pk: int,
        *,
        original: bool = False,
    ) -> DownloadedDocument:
        """Request exactly one resource item."""
        return await super().__call__(
            pk, RetrieveFileMode.DOWNLOAD, self._api_path, original=original
        )


class DocumentFilePreviewHelper(DocumentSubHelperBase):
    """Represent a factory for Paperless `DownloadedDocument` models."""

    _api_path = API_PATH["documents_preview"]

    async def __call__(  # type: ignore[override]
        self,
        pk: int,
        *,
        original: bool = False,
    ) -> DownloadedDocument:
        """Request exactly one resource item."""
        return await super().__call__(
            pk, RetrieveFileMode.PREVIEW, self._api_path, original=original
        )


class DocumentFileThumbnailHelper(DocumentSubHelperBase):
    """Represent a factory for Paperless `DownloadedDocument` models."""

    _api_path = API_PATH["documents_thumbnail"]

    async def __call__(  # type: ignore[override]
        self,
        pk: int,
        *,
        original: bool = False,
    ) -> DownloadedDocument:
        """Request exactly one resource item."""
        return await super().__call__(
            pk, RetrieveFileMode.THUMBNAIL, self._api_path, original=original
        )


class DocumentMetaHelper(HelperBase, helpers.CallableMixin[DocumentMeta]):
    """Represent a factory for Paperless `DocumentMeta` models."""

    _api_path = API_PATH["documents_meta"]
    _resource = PaperlessResource.DOCUMENTS

    _resource_cls = DocumentMeta


class DocumentNoteHelper(HelperBase):
    """Represent a factory for Paperless `DocumentNote` models."""

    _api_path = API_PATH["documents_notes"]
    _resource = PaperlessResource.DOCUMENTS

    _resource_cls = DocumentNote

    def __init__(self, api: "Paperless", attached_to: int | None = None) -> None:
        """Initialize a `DocumentHelper` instance."""
        super().__init__(api)

        self._attached_to = attached_to

    async def __call__(
        self,
        pk: int | None = None,
    ) -> list[DocumentNote]:
        """Request and return the documents `DocumentNote` list."""
        doc_pk = self._get_document_pk(pk)
        res = await self._api.request_json("get", self._get_api_path(doc_pk))

        # We have to transform data here slightly.
        # There are two major differences in the data depending on which endpoint is requested.
        # url: documents/{:pk}/ ->
        #       .document -> int
        #       .user -> int
        # url: documents/{:pk}/notes/ ->
        #       .document -> does not exist (so we add it here)
        #       .user -> dict(id=int, username=str, first_name=str, last_name=str)
        return [
            self._resource_cls.create_with_data(
                self._api,
                {
                    **item,
                    "document": doc_pk,
                    "user": item["user"]["id"] if self._api.host_api_version >= 8 else item["user"],
                },
                fetched=True,
            )
            for item in res
        ]

    def _get_document_pk(self, pk: int | None = None) -> int:
        """Return the attached document pk, or the parameter."""
        if not any((self._attached_to, pk)):
            message = f"Accessing {type(self).__name__} data without a primary key."
            raise PrimaryKeyRequiredError(message)
        return cast("int", self._attached_to or pk)

    def _get_api_path(self, pk: int) -> str:
        """Return the formatted api path."""
        return self._api_path.format(pk=pk)

    def draft(self, pk: int | None = None, **kwargs: Any) -> DocumentNoteDraft:
        """Return a fresh and empty `DocumentNoteDraft` instance.

        Example:
        -------
        ```python
        draft = paperless.documents.notes.draft(...)
        # do something
        ```

        """
        kwargs.update({"document": self._get_document_pk(pk)})
        return DocumentNoteDraft.create_with_data(
            self._api,
            data=kwargs,
            fetched=True,
        )


class DocumentHelper(
    HelperBase,
    helpers.SecurableMixin,
    helpers.CallableMixin[Document],
    helpers.DraftableMixin[DocumentDraft],
    helpers.IterableMixin[Document],
):
    """Represent a factory for Paperless `Document` models."""

    _api_path = API_PATH["documents"]
    _resource = PaperlessResource.DOCUMENTS

    _draft_cls = DocumentDraft
    _resource_cls = Document

    def __init__(self, api: "Paperless") -> None:
        """Initialize a `DocumentHelper` instance."""
        super().__init__(api)

        self._download = DocumentFileDownloadHelper(api)
        self._meta = DocumentMetaHelper(api)
        self._notes = DocumentNoteHelper(api)
        self._preview = DocumentFilePreviewHelper(api)
        self._suggestions = DocumentSuggestionsHelper(api)
        self._thumbnail = DocumentFileThumbnailHelper(api)

    @property
    def download(self) -> DocumentFileDownloadHelper:
        """Download the contents of an archived file.

        Example:
        -------
        ```python
        # request document contents directly...
        download = await paperless.documents.download(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)

        download = await doc.get_download()
        ```

        """
        return self._download

    @property
    def metadata(self) -> DocumentMetaHelper:
        """Return the attached `DocumentMetaHelper` instance.

        Example:
        -------
        ```python
        # request metadata of a document directly...
        metadata = await paperless.documents.metadata(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)
        metadata = await doc.get_metadata()
        ```

        """
        return self._meta

    @property
    def notes(self) -> DocumentNoteHelper:
        """Return the attached `DocumentNoteHelper` instance.

        Example:
        -------
        ```python
        # request document notes directly...
        notes = await paperless.documents.notes(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)
        notes = await doc.notes()
        ```

        """
        return self._notes

    @property
    def preview(self) -> DocumentFilePreviewHelper:
        """Preview the contents of an archived file.

        Example:
        -------
        ```python
        # request document contents directly...
        download = await paperless.documents.preview(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)

        download = await doc.get_preview()
        ```

        """
        return self._preview

    @property
    def suggestions(self) -> DocumentSuggestionsHelper:
        """Return the attached `DocumentSuggestionsHelper` instance.

        Example:
        -------
        ```python
        # request document suggestions directly...
        suggestions = await paperless.documents.suggestions(42)

        # ... or by using an already fetched document
        doc = await paperless.suggestions(42)

        suggestions = await doc.get_suggestions()
        ```

        """
        return self._suggestions

    @property
    def thumbnail(self) -> DocumentFileThumbnailHelper:
        """Download the contents of a thumbnail file.

        Example:
        -------
        ```python
        # request document contents directly...
        download = await paperless.documents.thumbnail(42)

        # ... or by using an already fetched document
        doc = await paperless.documents(42)

        download = await doc.get_thumbnail()
        ```

        """
        return self._thumbnail

    async def get_next_asn(self) -> int:
        """Request the next archive serial number from DRF."""
        async with self._api.request("get", API_PATH["documents_next_asn"]) as res:
            try:
                res.raise_for_status()
                return int(await res.text())
            except Exception as exc:
                raise AsnRequestError from exc

    async def more_like(self, pk: int) -> AsyncGenerator[Document]:
        """Lookup documents similar to the given document pk.

        Shortcut function. Same behaviour is possible using `reduce()`.

        Documentation: https://docs.paperless-ngx.com/api/#searching-for-documents
        """
        async with self.reduce(more_like_id=pk):
            async for item in self:
                yield item

    async def search(
        self, query: str | None = None, custom_field_query: str | None = None
    ) -> AsyncGenerator[Document]:
        """Lookup documents by a search query and/or custom_field_query.

        If none of both are provided, all documents are returned.

        Shortcut function. Same behaviour is possible using `reduce()`.

        Documentation: https://docs.paperless-ngx.com/usage/#basic-usage_searching
        """
        querykwargs = {}
        if query is not None:
            querykwargs["query"] = query
        if custom_field_query is not None:
            querykwargs["custom_field_query"] = custom_field_query

        async with self.reduce(**querykwargs):
            async for item in self:
                yield item

    async def email(
        self,
        documents: int | list[int],
        *,
        addresses: str,
        subject: str,
        message: str,
        use_archive_version: bool = True,
    ) -> None:
        """Email documents to one or more recipients as an attachment.

        Example:
        -------
        ```python
        # email document directly...
        await paperless.documents.email(
            [23, 42],
            addresses="example@example.com, another@example.com",
            subject="Subject",
            message="Message"
        )
        ```

        """
        data = {
            "documents": [documents] if isinstance(documents, int) else documents,
            "addresses": addresses,
            "subject": subject,
            "message": message,
            "use_archive_version": use_archive_version,
        }
        async with self._api.request("post", API_PATH["documents_email"], json=data) as res:
            try:
                res.raise_for_status()
            except Exception as exc:
                raise SendEmailError from exc



"""Paperless basic tests."""

import datetime
import re

import aiohttp
import pytest
from aioresponses import aioresponses

from pypaperless import Paperless
from pypaperless.const import API_PATH
from pypaperless.exceptions import (
    AsnRequestError,
    DraftFieldRequiredError,
    PrimaryKeyRequiredError,
    SendEmailError,
    TaskNotFoundError,
)
from pypaperless.models import (
    Config,
    CustomField,
    Document,
    DocumentDraft,
    DocumentMeta,
    DocumentNote,
    DocumentNoteDraft,
    Status,
    Task,
)
from pypaperless.models.common import (
    CUSTOM_FIELD_TYPE_VALUE_MAP,
    CustomFieldBooleanValue,
    CustomFieldDocumentLinkValue,
    CustomFieldValue,
    DocumentMetadataType,
    DocumentSearchHitType,
    RetrieveFileMode,
    StatisticDocumentFileTypeCount,
    StatusDatabaseType,
    StatusStorageType,
    StatusTasksType,
)
from pypaperless.models.documents import (
    DocumentCustomFieldList,
    DocumentSuggestions,
    DownloadedDocument,
)
from pypaperless.models.workflows import WorkflowActionHelper, WorkflowTriggerHelper

from . import DOCUMENT_MAP
from .const import PAPERLESS_TEST_URL
from .data import (
    DATA_CONFIG,
    DATA_CUSTOM_FIELDS,
    DATA_DOCUMENT_METADATA,
    DATA_DOCUMENT_NOTES,
    DATA_DOCUMENT_SUGGESTIONS,
    DATA_DOCUMENTS,
    DATA_DOCUMENTS_SEARCH,
    DATA_REMOTE_VERSION,
    DATA_STATISTICS,
    DATA_STATUS,
    DATA_TASKS,
)

# mypy: ignore-errors


# test models/config.py
class TestModelConfig:
    """Config test cases."""

    async def test_call(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test call."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['config_single']}".format(pk=1),
            status=200,
            payload=DATA_CONFIG[0],
        )
        item = await paperless.config(1)
        assert item
        assert isinstance(item, Config)
        # must raise as 1337 doesn't exist
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['config_single']}".format(pk=1337),
            status=404,
        )
        with pytest.raises(aiohttp.ClientResponseError):
            await paperless.config(1337)


# test models/documents.py
class TestModelDocuments:
    """Documents test cases."""

    async def test_lazy(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test laziness."""
        document = Document(paperless, data={"id": 1})
        assert not document.is_fetched

        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENTS["results"][0],
        )
        await document.load()
        assert document.is_fetched

    async def test_create(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test create."""
        defaults = DOCUMENT_MAP.draft_defaults or {}
        draft = paperless.documents.draft(**defaults)
        assert isinstance(draft, DocumentDraft)
        backup = draft.document
        draft.document = None
        with pytest.raises(DraftFieldRequiredError):
            await draft.save()
        draft.document = backup
        # actually call the create endpoint
        resp.post(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_post']}",
            status=200,
            payload="11112222-3333-4444-5555-666677778888",
        )
        await draft.save()

    async def test_create_date_property(self, paperless: Paperless) -> None:
        """Test create_date property - well, lol."""
        document = Document.create_with_data(
            paperless, data={**DATA_DOCUMENTS["results"][0]}, fetched=True
        )
        assert document.created_date == document.created

    async def test_udpate(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test update."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENTS["results"][0],
        )
        to_update = await paperless.documents(1)
        new_title = f"{to_update.title} Updated"
        to_update.title = new_title
        # actually call the update endpoint
        resp.patch(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload={
                **to_update._data,  # pylint: disable=protected-access
                "title": new_title,
            },
        )
        await to_update.update()
        assert to_update.title == new_title

    async def test_delete(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test delete."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENTS["results"][0],
        )
        to_delete = await paperless.documents(1)
        resp.delete(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=204,  # Paperless-ngx responds with 204 on deletion
        )
        assert await to_delete.delete()
        # test deletion failed
        resp.delete(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=404,  # we send another status code
        )
        assert not await to_delete.delete()

    async def test_meta(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test meta."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENTS["results"][0],
        )
        document = await paperless.documents(1)
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_meta']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENT_METADATA,
        )
        meta = await document.get_metadata()
        assert isinstance(meta, DocumentMeta)
        assert isinstance(meta.original_metadata, list)
        for item in meta.original_metadata:
            assert isinstance(item, DocumentMetadataType)
        assert isinstance(meta.archive_metadata, list)
        for item in meta.archive_metadata:
            assert isinstance(item, DocumentMetadataType)

    async def test_files(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test files."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENTS["results"][0],
        )
        document = await paperless.documents(1)
        resp.get(
            re.compile(
                r"^"
                + f"{PAPERLESS_TEST_URL}{API_PATH['documents_download']}".format(pk=1)
                + r"\?.*$"
            ),
            status=200,
            headers={
                "Content-Type": "application/pdf",
                "Content-Disposition": "attachment;filename=any_filename.pdf",
            },
            body=b"Binary data: download",
        )
        download = await document.get_download()
        assert isinstance(download, DownloadedDocument)
        assert download.mode == RetrieveFileMode.DOWNLOAD
        resp.get(
            re.compile(
                r"^"
                + f"{PAPERLESS_TEST_URL}{API_PATH['documents_preview']}".format(pk=1)
                + r"\?.*$"
            ),
            status=200,
            headers={
                "Content-Type": "application/pdf",
            },
            body=b"Binary data: preview",
        )
        preview = await document.get_preview()
        assert isinstance(preview, DownloadedDocument)
        assert preview.mode == RetrieveFileMode.PREVIEW
        resp.get(
            re.compile(
                r"^"
                + f"{PAPERLESS_TEST_URL}{API_PATH['documents_thumbnail']}".format(pk=1)
                + r"\?.*$"
            ),
            status=200,
            body=b"Binary data: thumbnail",
        )
        thumbnail = await document.get_thumbnail()
        assert isinstance(thumbnail, DownloadedDocument)
        assert thumbnail.mode == RetrieveFileMode.THUMBNAIL

    async def test_suggestions(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test suggestions."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENTS["results"][0],
        )
        document = await paperless.documents(1)
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_suggestions']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENT_SUGGESTIONS,
        )
        suggestions = await document.get_suggestions()
        assert isinstance(suggestions, DocumentSuggestions)

    async def test_get_next_an(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test get next asn."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_next_asn']}",
            status=200,
            payload=1337,
        )
        asn = await paperless.documents.get_next_asn()
        assert isinstance(asn, int)
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_next_asn']}",
            status=500,
        )
        with pytest.raises(AsnRequestError):
            await paperless.documents.get_next_asn()

    async def test_searching(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test searching."""
        # search
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH['documents']}" + r"\?.*query.*$"),
            status=200,
            payload=DATA_DOCUMENTS_SEARCH,
        )
        async for item in paperless.documents.search("1337"):
            assert isinstance(item, Document)
            assert item.has_search_hit
            assert isinstance(item.search_hit, DocumentSearchHitType)
        # custom_field_query
        resp.get(
            re.compile(
                r"^" + f"{PAPERLESS_TEST_URL}{API_PATH['documents']}" + r"\?.*custom_field_query.*$"
            ),
            status=200,
            payload=DATA_DOCUMENTS_SEARCH,
        )
        async for item in paperless.documents.search(custom_field_query="1337"):
            assert isinstance(item, Document)
            assert item.has_search_hit
            assert isinstance(item.search_hit, DocumentSearchHitType)
        # more_like
        resp.get(
            re.compile(
                r"^" + f"{PAPERLESS_TEST_URL}{API_PATH['documents']}" + r"\?.*more_like_id.*$"
            ),
            status=200,
            payload=DATA_DOCUMENTS_SEARCH,
        )
        async for item in paperless.documents.more_like(1337):
            assert isinstance(item, Document)
            assert item.has_search_hit
            assert isinstance(item.search_hit, DocumentSearchHitType)

    async def test_note_call(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test call."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENTS["results"][0],
        )
        item = await paperless.documents(1)
        assert isinstance(item, Document)
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_notes']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENT_NOTES,
        )
        results = await item.notes()
        assert isinstance(results, list)
        assert len(results) > 0
        for note in results:
            assert isinstance(note, DocumentNote)
            assert isinstance(note.created, datetime.datetime)
        with pytest.raises(PrimaryKeyRequiredError):
            item = await paperless.documents.notes()

    async def test_note_create(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test create."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENTS["results"][0],
        )
        item = await paperless.documents(1)
        draft = item.notes.draft(note="Test note.")
        assert isinstance(draft, DocumentNoteDraft)
        backup = draft.note
        draft.note = None
        with pytest.raises(DraftFieldRequiredError):
            await draft.save()
        draft.note = backup
        # actually call the create endpoint
        resp.post(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_notes']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENT_NOTES,
        )
        result = await draft.save()
        assert isinstance(result, tuple)

    async def test_note_delete(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test delete."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENTS["results"][0],
        )
        item = await paperless.documents(1)
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_notes']}".format(pk=1),
            status=200,
            payload=DATA_DOCUMENT_NOTES,
        )
        results = await item.notes()
        resp.delete(
            re.compile(
                r"^" + f"{PAPERLESS_TEST_URL}{API_PATH['documents_notes']}".format(pk=1) + r"\?.*$"
            ),
            status=204,  # Paperless-ngx responds with 204 on deletion
        )
        deletion = await results.pop().delete()
        assert deletion

    async def test_custom_field_list_wo_cache(
        self, resp: aioresponses, paperless: Paperless
    ) -> None:
        """Test custom field list without cache."""
        # request document
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=2),
            status=200,
            payload=DATA_DOCUMENTS["results"][1],
        )
        item = await paperless.documents(2)
        assert isinstance(item.custom_fields, DocumentCustomFieldList)

        # every item MUST NOT be a derived CustomFieldValue instance
        for field in item.custom_fields:
            for value_type in CUSTOM_FIELD_TYPE_VALUE_MAP.values():
                assert not isinstance(field, value_type)
            assert isinstance(field, CustomFieldValue)

    async def test_custom_field_list_wslash_cache(
        self, resp: aioresponses, paperless: Paperless
    ) -> None:
        """Test custom fields list with cache."""
        # set custom fields cache
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH['custom_fields']}" + r"\?.*$"),
            status=200,
            payload=DATA_CUSTOM_FIELDS,
        )
        paperless.cache.custom_fields = await paperless.custom_fields.as_dict()

        # request document
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_single']}".format(pk=2),
            status=200,
            payload=DATA_DOCUMENTS["results"][1],
        )
        item = await paperless.documents(2)
        assert isinstance(item.custom_fields, DocumentCustomFieldList)

        # every item may be a derived class or not
        for field in item.custom_fields:
            assert isinstance(field, CustomFieldValue)

        # test if custom field is in document custom field values
        test_cf = CustomField.create_with_data(
            api=paperless,
            data=DATA_CUSTOM_FIELDS["results"][0],
            fetched=True,
        )
        assert test_cf in item.custom_fields
        assert isinstance(item.custom_fields.get(test_cf), CustomFieldValue)
        assert item.custom_fields.default(test_cf) is not None
        assert item.custom_fields.default(-1337) is None

        # test typed getters
        assert isinstance(
            item.custom_fields.get(test_cf, CustomFieldDocumentLinkValue),
            CustomFieldDocumentLinkValue,
        )
        assert isinstance(
            item.custom_fields.default(test_cf, CustomFieldDocumentLinkValue),
            CustomFieldDocumentLinkValue,
        )
        with pytest.raises(TypeError):
            item.custom_fields.get(test_cf, CustomFieldBooleanValue)
        with pytest.raises(TypeError):
            item.custom_fields.default(test_cf, CustomFieldBooleanValue)

        # test remove field value
        item.custom_fields -= test_cf
        assert test_cf not in item.custom_fields

        # test add field value
        item.custom_fields += test_cf.draft_value(1337)
        assert test_cf in item.custom_fields

    async def test_email(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test sending emails."""
        # successful email sending
        resp.post(
            f"{PAPERLESS_TEST_URL}{API_PATH['documents_email']}",
            status=200,
            payload={"message": "Email sent"},
        )
        await paperless.documents.email(
            documents=1,
            addresses="test@example.org",
            subject="Test Email",
            message="Test Message",
        )

        # unsuccessful email sending
        resp.post(f"{PAPERLESS_TEST_URL}{API_PATH['documents_email']}", status=400)
        with pytest.raises(SendEmailError):
            await paperless.documents.email(
                documents=[1, 2],
                addresses="test@example.org",
                subject="Test Email",
                message="Test Message",
            )


# test models/remote_version.py
class TestModelVersion:
    """Version test cases."""

    async def test_call(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test call."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['remote_version']}",
            status=200,
            payload=DATA_REMOTE_VERSION,
        )
        remote_version = await paperless.remote_version()
        assert remote_version
        assert isinstance(remote_version.version, str)
        assert isinstance(remote_version.update_available, bool)


# test models/statistics.py
class TestModelStatistics:
    """Statistics test cases."""

    async def test_call(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test call."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['statistics']}",
            status=200,
            payload=DATA_STATISTICS,
        )
        stats = await paperless.statistics()
        assert stats
        assert isinstance(stats.character_count, int)
        assert isinstance(stats.document_file_type_counts, list)
        for item in stats.document_file_type_counts:
            assert isinstance(item, StatisticDocumentFileTypeCount)


# test models/status.py
class TestModelStatus:
    """Status test cases."""

    async def test_call(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test call."""
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['status']}",
            status=200,
            payload=DATA_STATUS,
        )
        status = await paperless.status()
        assert status
        assert isinstance(status, Status)
        assert isinstance(status.storage, StatusStorageType)
        assert isinstance(status.database, StatusDatabaseType)
        assert isinstance(status.tasks, StatusTasksType)

    async def test_has_errors(self, paperless: Paperless) -> None:
        """Test has errors."""
        data = {
            "database": {
                "status": "OK",
            },
            "tasks": {
                "redis_status": "OK",
                "celery_status": "OK",
                "classifier_status": "OK",
            },
        }

        # everything fine as we initialized Status with OK values only
        status = Status.create_with_data(paperless, data=data, fetched=True)
        assert status.has_errors is False

        # lets set something to ERROR
        data["database"]["status"] = "ERROR"
        status = Status.create_with_data(paperless, data=data, fetched=True)
        assert status.has_errors is True

        # assume any status value is None; None values are treated as no errors
        del data["database"]["status"]
        status = Status.create_with_data(paperless, data=data, fetched=True)
        assert status.has_errors is False


# test models/tasks.py
class TestModelTasks:
    """Tasks test cases."""

    async def test_iter(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test iter."""
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH['tasks']}" + r".*$"),
            status=200,
            payload=DATA_TASKS,
        )
        async for item in paperless.tasks:
            assert isinstance(item, Task)

    async def test_call(self, resp: aioresponses, paperless: Paperless) -> None:
        """Test call."""
        # by pk
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['tasks_single']}".format(pk=1),
            status=200,
            payload=DATA_TASKS[0],
        )
        item = await paperless.tasks(1)
        assert item
        assert isinstance(item, Task)
        # by uuid
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH['tasks']}" + r"\?task_id.*$"),
            status=200,
            payload=DATA_TASKS,
        )
        item = await paperless.tasks("dummy-found")
        assert item
        assert isinstance(item, Task)
        # must raise as pk doesn't exist
        resp.get(
            f"{PAPERLESS_TEST_URL}{API_PATH['tasks_single']}".format(pk=1337),
            status=404,
        )
        with pytest.raises(aiohttp.ClientResponseError):
            await paperless.tasks(1337)
        # must raise as task_id doesn't exist
        resp.get(
            re.compile(r"^" + f"{PAPERLESS_TEST_URL}{API_PATH['tasks']}" + r"\?task_id.*$"),
            status=200,
            payload=[],
        )
        with pytest.raises(TaskNotFoundError):
            await paperless.tasks("dummy-not-found")


# test models/workflows.py
class TestModelWorkflows:
    """Tasks test cases."""

    async def test_helpers(self, paperless: Paperless) -> None:
        """Test helpers."""
        assert isinstance(paperless.workflows.actions, WorkflowActionHelper)
        assert isinstance(paperless.workflows.triggers, WorkflowTriggerHelper)



"""Document types snapshot."""

DATA_DOCUMENT_TYPES = {
    "count": 5,
    "next": None,
    "previous": None,
    "all": [1, 2, 3, 4, 5],
    "results": [
        {
            "id": 1,
            "slug": "bescheid",
            "name": "Bescheid",
            "match": "",
            "matching_algorithm": 1,
            "is_insensitive": True,
            "document_count": 42,
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 2,
            "slug": "document",
            "name": "Document",
            "match": "",
            "matching_algorithm": 1,
            "is_insensitive": True,
            "document_count": 218,
            "owner": None,
            "user_can_change": None,
        },
        {
            "id": 3,
            "slug": "kontoauszug",
            "name": "Kontoauszug",
            "match": "",
            "matching_algorithm": 6,
            "is_insensitive": True,
            "document_count": 227,
            "owner": None,
            "user_can_change": None,
        },
        {
            "id": 4,
            "slug": "kundigung",
            "name": "Kuendigung",
            "match": "",
            "matching_algorithm": 1,
            "is_insensitive": True,
            "document_count": 24,
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 5,
            "slug": "rechnung",
            "name": "Rechnung",
            "match": "",
            "matching_algorithm": 6,
            "is_insensitive": True,
            "document_count": 246,
            "owner": None,
            "user_can_change": True,
        },
    ],
}



"""Mail snapshots."""

# mypy: ignore-errors

DATA_MAIL_ACCOUNTS = {
    "count": 1,
    "next": None,
    "previous": None,
    "all": [1],
    "results": [
        {
            "id": 1,
            "name": "Test Account",
            "imap_server": "imap.omega.net",
            "imap_port": 1337,
            "imap_security": 2,
            "username": "omega-weapon",
            "password": "********************",
            "character_set": "UTF-8",
            "is_token": False,
            "owner": 1,
            "user_can_change": True,
        }
    ],
}

DATA_MAIL_RULES = {
    "count": 1,
    "next": None,
    "previous": None,
    "all": [1],
    "results": [
        {
            "id": 1,
            "name": "Test",
            "account": 1,
            "folder": "INBOX",
            "filter_from": None,
            "filter_to": None,
            "filter_subject": None,
            "filter_body": None,
            "filter_attachment_filename_include": None,
            "filter_attachment_filename_exclude": None,
            "maximum_age": 3,
            "action": 3,
            "action_parameter": None,
            "assign_title_from": 1,
            "assign_tags": [],
            "assign_correspondent_from": 1,
            "assign_correspondent": None,
            "assign_document_type": None,
            "assign_owner_from_rule": True,
            "order": 1,
            "attachment_type": 1,
            "consumption_scope": 1,
            "owner": 1,
            "user_can_change": True,
        }
    ],
}

DATA_PROCESSED_MAIL = {
    "count": 3,
    "next": None,
    "previous": None,
    "all": [1, 2, 3],
    "results": [
        {
            "id": 1,
            "owner": 1,
            "rule": 1,
            "folder": "INBOX",
            "uid": "abcdef",
            "subject": "Invoice #1",
            "received": "2025-11-16T08:18:25+01:00",
            "processed": "2025-11-16T11:00:17.838202+01:00",
            "status": "SUCCESS",
            "error": None,
        },
        {
            "id": 2,
            "owner": 1,
            "rule": 1,
            "folder": "INBOX",
            "uid": "fedcba",
            "subject": "Invoice #2",
            "received": "2025-11-16T07:56:16+01:00",
            "processed": "2025-11-16T08:00:17.660439+01:00",
            "status": "SUCCESS",
            "error": None,
        },
        {
            "id": 3,
            "owner": 1,
            "rule": 1,
            "folder": "INBOX",
            "uid": "123456",
            "subject": "Invoice #3",
            "received": "2025-11-16T13:54:23+01:00",
            "processed": "2025-11-16T14:00:17.846562+01:00",
            "status": "SUCCESS",
            "error": None,
        },
    ],
}



"""Statistics snapshot."""

DATA_STATISTICS = {
    "documents_total": 1337,
    "documents_inbox": 2,
    "inbox_tag": 1,
    "inbox_tags": [1],
    "document_file_type_counts": [
        {"mime_type": "application/pdf", "mime_type_count": 1334},
        {"mime_type": "image/jpeg", "mime_type_count": 2},
        {"mime_type": "message/rfc822", "mime_type_count": 1},
    ],
    "character_count": 13371337,
    "tag_count": 5,
    "correspondent_count": 42,
    "document_type_count": 23,
    "storage_path_count": 5,
    "current_asn": 84,
}



"""Documents search snapshot."""

# mypy: ignore-errors

DATA_DOCUMENTS_SEARCH = {
    "count": 1,
    "next": None,
    "previous": None,
    "all": [1],
    "results": [
        {
            "id": 1,
            "correspondent": 1,
            "document_type": 2,
            "storage_path": None,
            "title": "Crazy Document",
            "content": "some OCRd text",
            "tags": [],
            "created": "2011-06-22T00:00:00+00:00",
            "created_date": "2011-06-22",
            "modified": "2023-08-08T06:06:35.495972+00:00",
            "added": "2023-06-30T05:44:14.317925+00:00",
            "archive_serial_number": None,
            "original_file_name": "Scan_2023-06-29_113857.pdf",
            "archived_file_name": "2011-06-22 filename.pdf",
            "owner": 2,
            "user_can_change": True,
            "notes": [],
            "custom_fields": [],
            "__search_hit__": {
                "score": 1.0,
                "highlights": "some neat hint",
                "note_highlights": "",
                "rank": 0,
            },
        },
    ],
}



"""Document metadata snapshot."""

DATA_DOCUMENT_METADATA = {
    "original_checksum": "18e2352cc13379d19bd9ce329428bb99",
    "original_size": 190348,
    "original_mime_type": "application/pdf",
    "media_filename": "xxx.pdf",
    "has_archive_version": True,
    "original_metadata": [],
    "archive_checksum": "8a69542583571337dd263f8bb3cf23bd",
    "archive_media_filename": "0000226.pdf",
    "original_filename": None,
    "lang": "de",
    "archive_size": 161371,
    "archive_metadata": [
        {
            "namespace": "http://ns.adobe.com/pdf/1.3/",
            "prefix": "pdf",
            "key": "Producer",
            "value": "pikepdf 2.16.1",
        },
        {
            "namespace": "http://ns.adobe.com/xap/1.0/",
            "prefix": "xmp",
            "key": "ModifyDate",
            "value": "2022-02-12T11:34:40+00:00",
        },
        {
            "namespace": "http://ns.adobe.com/xap/1.0/",
            "prefix": "xmp",
            "key": "CreateDate",
            "value": "2022-01-07T13:40:49",
        },
        {
            "namespace": "http://ns.adobe.com/xap/1.0/",
            "prefix": "xmp",
            "key": "CreatorTool",
            "value": "ocrmypdf 12.3.2 / Tesseract OCR-PDF 4.1.1",
        },
        {
            "namespace": "http://ns.adobe.com/xap/1.0/mm/",
            "prefix": "xmpMM",
            "key": "DocumentID",
            "value": "uuid:13371337-2342-1337-4242-133723426665",
        },
        {
            "namespace": "http://purl.org/dc/elements/1.1/",
            "prefix": "dc",
            "key": "format",
            "value": "application/pdf",
        },
        {
            "namespace": "http://purl.org/dc/elements/1.1/",
            "prefix": "dc",
            "key": "title",
            "value": "Octopus Energy updated",
        },
        {
            "namespace": "http://www.aiim.org/pdfa/ns/id/",
            "prefix": "pdfaid",
            "key": "part",
            "value": "2",
        },
        {
            "namespace": "http://www.aiim.org/pdfa/ns/id/",
            "prefix": "pdfaid",
            "key": "conformance",
            "value": "B",
        },
        {
            "namespace": "http://purl.org/dc/elements/1.1/",
            "prefix": "dc",
            "key": "creator",
            "value": "None",
        },
        {
            "namespace": "http://ns.adobe.com/xap/1.0/",
            "prefix": "xmp",
            "key": "MetadataDate",
            "value": "2022-02-12T11:34:40.383718+00:00",
        },
    ],
}



"""Tasks snapshot."""

DATA_TASKS = [
    {
        "id": 1,
        "task_id": "11112222-aaaa-bbbb-cccc-333344445555",
        "task_file_name": "a.png",
        "date_created": "2023-12-16T13:06:29.107815+00:00",
        "date_done": None,
        "type": "file",
        "status": "STARTED",
        "result": None,
        "acknowledged": False,
        "related_document": None,
    },
    {
        "id": 2,
        "task_id": "ffffeeee-9999-8888-7777-ddddccccbbbb",
        "task_file_name": "b.png",
        "date_created": "2023-12-16T13:06:26.117158+00:00",
        "date_done": "2023-12-16T13:06:29.859669+00:00",
        "type": "file",
        "status": "SUCCESS",
        "result": "Success. New document id 1780 created",
        "acknowledged": False,
        "related_document": "1780",
    },
    {
        "id": 3,
        "task_id": "abcdef12-3456-7890-abcd-ef1234567890",
        "task_file_name": "c.png",
        "date_created": "2023-12-16T13:04:28.175624+00:00",
        "date_done": "2023-12-16T13:04:32.318797+00:00",
        "type": "file",
        "status": "SUCCESS",
        "result": "Success. New document id 1779 created",
        "acknowledged": False,
        "related_document": "1779",
    },
]



"""Config snapshot."""

DATA_CONFIG = [
    {
        "id": 1,
        "user_args": None,
        "output_type": "pdf",
        "pages": None,
        "language": "eng",
        "mode": None,
        "skip_archive_file": None,
        "image_dpi": None,
        "unpaper_clean": None,
        "deskew": None,
        "rotate_pages": None,
        "rotate_pages_threshold": None,
        "max_image_pixels": None,
        "color_conversion_strategy": None,
        "app_title": None,
        "app_logo": None,
    }
]



"""Raw data constants."""

import json
from pathlib import Path

from .config import DATA_CONFIG
from .correspondents import DATA_CORRESPONDENTS
from .custom_fields import DATA_CUSTOM_FIELDS
from .document_metadata import DATA_DOCUMENT_METADATA
from .document_notes import DATA_DOCUMENT_NOTES
from .document_suggestions import DATA_DOCUMENT_SUGGESTIONS
from .document_types import DATA_DOCUMENT_TYPES
from .documents import DATA_DOCUMENTS
from .documents_search import DATA_DOCUMENTS_SEARCH
from .groups import DATA_GROUPS
from .mail import DATA_MAIL_ACCOUNTS, DATA_MAIL_RULES, DATA_PROCESSED_MAIL
from .object_permissions import DATA_OBJECT_PERMISSIONS
from .paths import DATA_PATHS
from .remote_version import DATA_REMOTE_VERSION
from .saved_views import DATA_SAVED_VIEWS
from .share_links import DATA_SHARE_LINKS
from .statistics import DATA_STATISTICS
from .status import DATA_STATUS
from .storage_paths import DATA_STORAGE_PATHS
from .tags import DATA_TAGS
from .tasks import DATA_TASKS
from .token import DATA_TOKEN
from .users import DATA_USERS
from .workflow import DATA_WORKFLOW_ACTIONS, DATA_WORKFLOW_TRIGGERS, DATA_WORKFLOWS

# mypy: ignore-errors


def _read_schema() -> dict:
    filepath = Path("tests/data/schema.json")
    with Path.open(filepath, mode="r", encoding="utf-8") as file:
        return json.load(file)


DATA_SCHEMA = _read_schema()


__all__ = (
    "DATA_CONFIG",
    "DATA_CORRESPONDENTS",
    "DATA_CUSTOM_FIELDS",
    "DATA_DOCUMENTS",
    "DATA_DOCUMENTS_SEARCH",
    "DATA_DOCUMENT_METADATA",
    "DATA_DOCUMENT_NOTES",
    "DATA_DOCUMENT_SUGGESTIONS",
    "DATA_DOCUMENT_TYPES",
    "DATA_GROUPS",
    "DATA_MAIL_ACCOUNTS",
    "DATA_MAIL_RULES",
    "DATA_OBJECT_PERMISSIONS",
    "DATA_PATHS",
    "DATA_PROCESSED_MAIL",
    "DATA_REMOTE_VERSION",
    "DATA_SAVED_VIEWS",
    "DATA_SCHEMA",
    "DATA_SHARE_LINKS",
    "DATA_STATISTICS",
    "DATA_STATUS",
    "DATA_STORAGE_PATHS",
    "DATA_TAGS",
    "DATA_TASKS",
    "DATA_TOKEN",
    "DATA_USERS",
    "DATA_WORKFLOWS",
    "DATA_WORKFLOW_ACTIONS",
    "DATA_WORKFLOW_TRIGGERS",
)



"""Correspondents snapshot."""

DATA_CORRESPONDENTS = {
    "count": 5,
    "next": None,
    "previous": None,
    "all": [1, 2, 3, 4, 5],
    "results": [
        {
            "id": 1,
            "slug": "sample-correspondent",
            "name": "Sample Correspondent",
            "match": "",
            "matching_algorithm": 1,
            "is_insensitive": True,
            "document_count": 12,
            "last_correspondence": "2022-10-18",
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 2,
            "slug": "burger-fastfood-delivery-billing",
            "name": "Burger FastFood Delivery Billing",
            "match": "",
            "matching_algorithm": 1,
            "is_insensitive": True,
            "document_count": 6,
            "last_correspondence": "2006-03-07",
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 3,
            "slug": "example-company",
            "name": "Example Company",
            "match": "",
            "matching_algorithm": 6,
            "is_insensitive": True,
            "document_count": 3,
            "last_correspondence": "2022-03-11",
            "owner": 3,
            "user_can_change": True,
        },
        {
            "id": 4,
            "slug": "no-illness-more-money-health-insurance",
            "name": "No Illness More Money Health Insurance",
            "match": "",
            "matching_algorithm": 1,
            "is_insensitive": True,
            "document_count": 6,
            "last_correspondence": "2000-07-18",
            "owner": None,
            "user_can_change": True,
        },
        {
            "id": 5,
            "slug": "your-cash-is-my-cash-bank",
            "name": "Your Cash is my Cash Bank",
            "match": "Your Cash My Bank",
            "matching_algorithm": 2,
            "is_insensitive": True,
            "document_count": 20,
            "last_correspondence": "2023-08-28",
            "owner": None,
            "user_can_change": True,
        },
    ],
}



"""Groups snapshot."""

DATA_GROUPS = {
    "count": 1,
    "next": None,
    "previous": None,
    "all": [1],
    "results": [
        {
            "id": 1,
            "name": "Sample Group",
            "permissions": [
                "add_consumptiontemplate",
                "change_consumptiontemplate",
                "delete_consumptiontemplate",
                "view_consumptiontemplate",
                "view_correspondent",
                "add_customfield",
                "change_customfield",
                "delete_customfield",
                "view_customfield",
                "add_customfieldinstance",
                "change_customfieldinstance",
                "delete_customfieldinstance",
                "view_customfieldinstance",
                "add_document",
                "change_document",
                "delete_document",
                "view_document",
                "view_documenttype",
                "add_note",
                "change_note",
                "delete_note",
                "view_note",
                "add_savedview",
                "change_savedview",
                "delete_savedview",
                "view_savedview",
                "add_sharelink",
                "change_sharelink",
                "delete_sharelink",
                "view_sharelink",
                "view_tag",
                "add_uisettings",
                "change_uisettings",
                "delete_uisettings",
                "view_uisettings",
            ],
        },
    ],
}
